# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 08.05.2023

  Purpose: Raise class for formatting thrown exception messages.
  The message can be formatted with information about the class,
  method, and line number where the exception was thrown.
"""
from types import FrameType
from typing import Optional
from jsktoolbox.attribtool import NoDynamicAttributes


class Raise(NoDynamicAttributes):
    """Raise class for formatting thrown exception messages."""

    @classmethod
    def message(
        cls,
        message: str,
        class_name: str = "",
        currentframe: Optional[FrameType] = None,
    ) -> str:
        """Message formatter method.

        message: str    - message to format
        class_name: str - caller class name (self.__class__.__name__)
        currentframe: FrameType - object from inspect.currentframe()

        Return: formatted message string
        """
        template: str = f"{message}"
        if currentframe and isinstance(currentframe, FrameType):
            template = f"{currentframe.f_code.co_name} [line:{currentframe.f_lineno}]: {template}"
        elif isinstance(class_name, str) and class_name != "":
            template = f"{class_name}: {template}"
            return template
        else:
            return template
        template = f"{class_name}.{template}"
        return template

    @classmethod
    def error(
        cls,
        message: str,
        exception: type[Exception] = Exception,
        class_name: str = "",
        currentframe: Optional[FrameType] = None,
    ) -> Exception:
        """Return exception with formatted string.

        message: str - message to format
        exception: type[Exception] - custom exception to return
        class_name: str - caller class name (self.__class__.__name__)
        currentframe: FrameType - object from inspect.currentframe()

        Return: given exception type"""
        if isinstance(exception, type):
            if not isinstance(exception(), Exception):
                raise cls.error(
                    f"Exception class or its derived class expected, '{exception.__qualname__}' received.",
                    TypeError,
                    class_name,
                    currentframe,
                )
        else:
            raise cls.error(
                "Exception class or its derived class expected.",
                TypeError,
                class_name,
                currentframe,
            )
        return exception(
            cls.message(
                f"[{exception.__qualname__}]: {message}"
                if message
                else f"[{exception.__qualname__}]",
                class_name,
                currentframe,
            )
        )


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 03.11.2023

  Purpose: Base class for classess derived from threading.Thread
"""

from inspect import currentframe
from typing import Any, Optional, Tuple, Dict
from threading import Event
from jsktoolbox.libs.base_data import BData
from jsktoolbox.attribtool import NoDynamicAttributes, ReadOnlyClass
from jsktoolbox.raisetool import Raise


class _Keys(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    ARGS = "_args"
    DAEMONIC = "_daemonic"
    DEBUG = "_debug"
    IDENT = "_ident"
    INVOKE_EXCEPTHOOK = "_invoke_excepthook"
    IS_STOPPED = "_is_stopped"
    KWARGS = "_kwargs"
    NAME = "_name"
    NATIVE_ID = "_native_id"
    SLEEP_PERIOD = "_sleep_period"
    STARTED = "_started"
    STDERR = "_stderr"
    STOP_EVENT = "_stop_event"
    TARGET = "_target"
    TSTATE_LOCK = "_tstate_lock"


class ThBaseObject(BData, NoDynamicAttributes):
    """Base class for classes derived from threading.Thread.

    Definition of properties used in the threading library.
    """

    @property
    def _target(self) -> Optional[Any]:
        if _Keys.TARGET not in self._data:
            self._data[_Keys.TARGET] = None
        return self._data[_Keys.TARGET]

    @_target.setter
    def _target(self, value: Any) -> None:
        self._data[_Keys.TARGET] = value

    @property
    def _name(self) -> Optional[str]:
        if _Keys.NAME not in self._data:
            self._data[_Keys.NAME] = None
        return self._data[_Keys.NAME]

    @_name.setter
    def _name(self, value: Optional[str]) -> None:
        if value is not None and not isinstance(value, str):
            raise Raise.error(
                f"Expected String type, received: '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.NAME] = value

    @property
    def _args(self) -> Optional[Tuple]:
        if _Keys.ARGS not in self._data:
            self._data[_Keys.ARGS] = None
        return self._data[_Keys.ARGS]

    @_args.setter
    def _args(self, value: Tuple) -> None:
        self._data[_Keys.ARGS] = value

    @property
    def _kwargs(self) -> Optional[Dict]:
        if _Keys.KWARGS not in self._data:
            self._data[_Keys.KWARGS] = None
        return self._data[_Keys.KWARGS]

    @_kwargs.setter
    def _kwargs(self, value: Dict) -> None:
        if value is not None and not isinstance(value, Dict):
            raise Raise.error(
                f"Expected Dict type, received: '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.KWARGS] = value

    @property
    def _daemonic(self) -> Optional[bool]:
        if _Keys.DAEMONIC not in self._data:
            self._data[_Keys.DAEMONIC] = None
        return self._data[_Keys.DAEMONIC]

    @_daemonic.setter
    def _daemonic(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise Raise.error(
                f"Expected Boolean type, received: '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.DAEMONIC] = value

    @property
    def _debug(self) -> Optional[bool]:
        if _Keys.DEBUG not in self._data:
            self._data[_Keys.DEBUG] = None
        return self._data[_Keys.DEBUG]

    @_debug.setter
    def _debug(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise Raise.error(
                f"Expected Boolean type, received: '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.DEBUG] = value

    @property
    def _ident(self) -> Optional[int]:
        if _Keys.IDENT not in self._data:
            self._data[_Keys.IDENT] = None
        return self._data[_Keys.IDENT]

    @_ident.setter
    def _ident(self, value: Optional[int]) -> None:
        if value is not None and not isinstance(value, int):
            raise Raise.error(
                f"Expected Integer type, received: '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.IDENT] = value

    @property
    def _native_id(self) -> Optional[int]:
        if _Keys.NATIVE_ID not in self._data:
            self._data[_Keys.NATIVE_ID] = None
        return self._data[_Keys.NATIVE_ID]

    @_native_id.setter
    def _native_id(self, value: Optional[int]) -> None:
        if value is not None and not isinstance(value, int):
            raise Raise.error(
                f"Expected Integer type, received '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.NATIVE_ID] = value

    @property
    def _tstate_lock(self) -> Optional[Any]:
        if _Keys.TSTATE_LOCK not in self._data:
            self._data[_Keys.TSTATE_LOCK] = None
        return self._data[_Keys.TSTATE_LOCK]

    @_tstate_lock.setter
    def _tstate_lock(self, value: Any) -> None:
        self._data[_Keys.TSTATE_LOCK] = value

    @property
    def _started(self) -> Optional[Event]:
        if _Keys.STARTED not in self._data:
            self._data[_Keys.STARTED] = None
        return self._data[_Keys.STARTED]

    @_started.setter
    def _started(self, value: Event) -> None:
        if value is not None and not isinstance(value, Event):
            raise Raise.error(
                f"Expected threading.Event type, received: '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.STARTED] = value

    @property
    def _is_stopped(self) -> Optional[bool]:
        if _Keys.IS_STOPPED not in self._data:
            self._data[_Keys.IS_STOPPED] = None
        return self._data[_Keys.IS_STOPPED]

    @_is_stopped.setter
    def _is_stopped(self, value: bool) -> None:
        if not isinstance(value, bool):
            raise Raise.error(
                f"Expected Boolean type, received: '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.IS_STOPPED] = value

    @property
    def _stderr(self) -> Optional[Any]:
        if _Keys.STDERR not in self._data:
            self._data[_Keys.STDERR] = None
        return self._data[_Keys.STDERR]

    @_stderr.setter
    def _stderr(self, value: Any) -> None:
        self._data[_Keys.STDERR] = value

    @property
    def _invoke_excepthook(self) -> Optional[Any]:
        if _Keys.INVOKE_EXCEPTHOOK not in self._data:
            self._data[_Keys.INVOKE_EXCEPTHOOK] = None
        return self._data[_Keys.INVOKE_EXCEPTHOOK]

    @_invoke_excepthook.setter
    def _invoke_excepthook(self, value: Any) -> None:
        self._data[_Keys.INVOKE_EXCEPTHOOK] = value

    @property
    def _stop_event(self) -> Optional[Event]:
        if _Keys.STOP_EVENT not in self._data:
            self._data[_Keys.STOP_EVENT] = None
        return self._data[_Keys.STOP_EVENT]

    @_stop_event.setter
    def _stop_event(self, obj: Event) -> None:
        if obj is not None and not isinstance(obj, Event):
            raise Raise.error(
                f"Expected threading.Event type, received: '{type(obj)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.STOP_EVENT] = obj

    @property
    def is_stopped(self) -> Optional[bool]:
        return self._is_stopped

    @property
    def started(self) -> bool:
        if self._started is not None:
            return self._started.is_set()
        return False

    @property
    def sleep_period(self) -> float:
        """Return sleepperiod value."""
        if _Keys.SLEEP_PERIOD not in self._data:
            self._data[_Keys.SLEEP_PERIOD] = 1.0
        return self._data[_Keys.SLEEP_PERIOD]

    @sleep_period.setter
    def sleep_period(self, value: float) -> None:
        """Set sleepperiod value."""
        if not isinstance(value, float):
            raise Raise.error(
                f"Expected positive float type, received: '{value}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.SLEEP_PERIOD] = value


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 10.10.2023

  Purpose: logger engine classes.
"""

import os
import sys
import syslog

from inspect import currentframe
from typing import Optional, Union

from jsktoolbox.attribtool import NoDynamicAttributes
from jsktoolbox.raisetool import Raise
from jsktoolbox.libs.base_data import BData
from jsktoolbox.libs.system import PathChecker
from jsktoolbox.libs.base_logs import (
    BLoggerEngine,
    Keys,
    SysLogKeys,
)
from jsktoolbox.libs.interfaces.logger_engine import ILoggerEngine
from jsktoolbox.logstool.formatters import BLogFormatter

# https://www.geeksforgeeks.org/python-testing-output-to-stdout/


class LoggerEngineStdout(ILoggerEngine, BLoggerEngine, BData, NoDynamicAttributes):
    """STDOUT Logger engine."""

    def __init__(
        self,
        name: Optional[str] = None,
        formatter: Optional[BLogFormatter] = None,
        buffered: bool = False,
    ) -> None:
        """Constructor."""
        if name is not None:
            self.name = name
        self._data[Keys.BUFFERED] = buffered
        self._data[Keys.FORMATTER] = None
        if formatter is not None:
            if isinstance(formatter, BLogFormatter):
                self._data[Keys.FORMATTER] = formatter
            else:
                raise Raise.error(
                    f"Expected LogFormatter type, received: '{type(formatter)}'.",
                    TypeError,
                    self._c_name,
                    currentframe(),
                )

    def send(self, message: str) -> None:
        """Send message to STDOUT."""
        if self._data[Keys.FORMATTER]:
            message = self._data[Keys.FORMATTER].format(message, self.name)
        sys.stdout.write(f"{message}")
        if not f"{message}".endswith("\n"):
            sys.stdout.write("\n")
        if not self._data[Keys.BUFFERED]:
            sys.stdout.flush()


class LoggerEngineStderr(ILoggerEngine, BLoggerEngine, BData, NoDynamicAttributes):
    """STDERR Logger engine."""

    def __init__(
        self,
        name: Optional[str] = None,
        formatter: Optional[BLogFormatter] = None,
        buffered: bool = False,
    ) -> None:
        """Constructor."""
        if name is not None:
            self.name = name
        self._data[Keys.BUFFERED] = buffered
        self._data[Keys.FORMATTER] = None
        if formatter is not None:
            if isinstance(formatter, BLogFormatter):
                self._data[Keys.FORMATTER] = formatter
            else:
                raise Raise.error(
                    f"Expected LogFormatter type, received: '{type(formatter)}'.",
                    TypeError,
                    self._c_name,
                    currentframe(),
                )

    def send(self, message: str) -> None:
        """Send message to STDERR."""
        if self._data[Keys.FORMATTER]:
            message = self._data[Keys.FORMATTER].format(message, self.name)
        sys.stderr.write(f"{message}")
        if not f"{message}".endswith("\n"):
            sys.stderr.write("\n")
        if not self._data[Keys.BUFFERED]:
            sys.stderr.flush()


class LoggerEngineFile(ILoggerEngine, BLoggerEngine, BData, NoDynamicAttributes):
    """FILE Logger engine."""

    def __init__(
        self,
        name: Optional[str] = None,
        formatter: Optional[BLogFormatter] = None,
        buffered: bool = False,
    ) -> None:
        """Constructor."""
        if name is not None:
            self.name = name
        self._data[Keys.BUFFERED] = buffered
        self._data[Keys.FORMATTER] = None
        if formatter is not None:
            if isinstance(formatter, BLogFormatter):
                self._data[Keys.FORMATTER] = formatter
            else:
                raise Raise.error(
                    f"Expected LogFormatter type, received: '{type(formatter)}'.",
                    TypeError,
                    self._c_name,
                    currentframe(),
                )

    def send(self, message: str) -> None:
        """Send message to file."""
        if self._data[Keys.FORMATTER]:
            message = self._data[Keys.FORMATTER].format(message, self.name)
            if self.logfile is None:
                raise Raise.error(
                    f"The {self._c_name} is not configured correctly.",
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
            logdir: str = self.logdir if self.logdir else ""
            with open(os.path.join(logdir, self.logfile), "a") as file:
                if file.writable:
                    file.write(message)
                    file.write("\n")

    @property
    def logdir(self) -> Optional[str]:
        """Return log directory."""
        if Keys.DIR not in self._data:
            self._data[Keys.DIR] = None
        return self._data[Keys.DIR]

    @logdir.setter
    def logdir(self, dirname: str) -> None:
        """Set log directory."""
        if dirname[-1] != os.sep:
            dirname = f"{dirname}/"
        ld = PathChecker(dirname)
        if not ld.exists:
            ld.create()
        if ld.exists and ld.is_dir:
            self._data[Keys.DIR] = ld.path

    @property
    def logfile(self) -> Optional[str]:
        """Return log file name."""
        if Keys.FILE not in self._data:
            self._data[Keys.FILE] = None
        return self._data[Keys.FILE]

    @logfile.setter
    def logfile(self, filename: str) -> None:
        """Set log file name."""
        # TODO: check procedure
        fn = None
        if self.logdir is None:
            fn = filename
        else:
            fn = os.path.join(self.logdir, filename)
        ld = PathChecker(fn)
        if ld.exists:
            if not ld.is_file:
                raise Raise.error(
                    f"The 'filename' passed: '{filename}', is a directory.",
                    FileExistsError,
                    self._c_name,
                    currentframe(),
                )
        else:
            if not ld.create():
                raise Raise.error(
                    f"I cannot create a file: {ld.path}",
                    PermissionError,
                    self._c_name,
                    currentframe(),
                )
        self.logdir = ld.dirname if ld.dirname else ""
        self._data[Keys.FILE] = ld.filename


class LoggerEngineSyslog(ILoggerEngine, BLoggerEngine, BData, NoDynamicAttributes):
    """SYSLOG Logger engine."""

    def __init__(
        self,
        name: Optional[str] = None,
        formatter: Optional[BLogFormatter] = None,
        buffered: bool = False,
    ) -> None:
        """Constructor."""
        if name is not None:
            self.name = name
        self._data[Keys.BUFFERED] = buffered
        self._data[Keys.FORMATTER] = None
        self._data[Keys.LEVEL] = SysLogKeys.level.INFO
        self._data[Keys.FACILITY] = SysLogKeys.facility.USER
        self._data[Keys.SYSLOG] = None
        if formatter is not None:
            if isinstance(formatter, BLogFormatter):
                self._data[Keys.FORMATTER] = formatter
            else:
                raise Raise.error(
                    f"Expected LogFormatter type, received: '{type(formatter)}'.",
                    TypeError,
                    self._c_name,
                    currentframe(),
                )

    def __del__(self) -> None:
        try:
            self._data[Keys.SYSLOG].closelog()
        except:
            pass
        self._data[Keys.SYSLOG] = None

    @property
    def facility(self) -> int:
        """Return syslog facility."""
        return self._data[Keys.FACILITY]

    @facility.setter
    def facility(self, value: Union[int, str]) -> None:
        """Set syslog facility."""
        if isinstance(value, int):
            if value in tuple(SysLogKeys.facility_keys.values()):
                self._data[Keys.FACILITY] = value
            else:
                raise Raise.error(
                    f"Syslog facility: '{value}' not found.",
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
        if isinstance(value, str):
            if value in SysLogKeys.facility_keys:
                self._data[Keys.FACILITY] = SysLogKeys.facility_keys[value]
            else:
                raise Raise.error(
                    f"Syslog facility name not found: '{value}'",
                    KeyError,
                    self._c_name,
                    currentframe(),
                )
        try:
            self._data[Keys.SYSLOG].closelog()
        except:
            pass
        self._data[Keys.SYSLOG] = None

    @property
    def level(self) -> int:
        """Return syslog level."""
        return self._data[Keys.LEVEL]

    @level.setter
    def level(self, value: Union[int, str]) -> None:
        """Set syslog level."""
        if isinstance(value, int):
            if value in tuple(SysLogKeys.level_keys.values()):
                self._data[Keys.LEVEL] = value
            else:
                raise Raise.error(
                    f"Syslog level: '{value}' not found.",
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
        if isinstance(value, str):
            if value in SysLogKeys.level_keys:
                self._data[Keys.LEVEL] = SysLogKeys.level_keys[value]
            else:
                raise Raise.error(
                    f"Syslog level name not found: '{value}'",
                    KeyError,
                    self._c_name,
                    currentframe(),
                )
        try:
            self._data[Keys.SYSLOG].closelog()
        except:
            pass
        self._data[Keys.SYSLOG] = None

    def send(self, message: str) -> None:
        """Send message to SYSLOG."""
        if self._data[Keys.FORMATTER]:
            message = self._data[Keys.FORMATTER].format(message, self.name)
        if self._data[Keys.SYSLOG] is None:
            self._data[Keys.SYSLOG] = syslog
            self._data[Keys.SYSLOG].openlog(facility=self._data[Keys.FACILITY])
        self._data[Keys.SYSLOG].syslog(priority=self._data[Keys.LEVEL], message=message)


# #[EOF]#######################################################################



# -*- coding: utf-8 -*-
"""
  widgets.py
  Author : Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
  Created: 15.01.2024, 10:42:01
  
  Purpose: custom tkinter widgets.

  VerticalScrolledFrame: https://gist.github.com/novel-yet-trivial/3eddfce704db3082e38c84664fc1fdf8
"""


import tkinter as tk
from tkinter import Toplevel, ttk
from typing import Any, Optional, List, Tuple, Union, Dict

from jsktoolbox.tktool.base import TkBase


class CreateToolTip(TkBase):
    """Create a tooltip for a given widget.

    Constructor:
    widget: tk.Misc -- Parent widget handler,
    text: Union[str, List[str], Tuple[str], tk.StringVar] -- text displayed in tooltip,
    wait_time: int -- delay of displaying tooltip [ms],
    wrap_length: int -- Limit the number of characters on each line to the specified value.
                    The default value of 0 means that lines will only be broken on newlines.
    """

    __id: Optional[str] = None
    __tw: Optional[tk.Toplevel] = None
    __waittime: int = None  # type: ignore
    __widget: tk.Misc = None  # type: ignore
    __wraplength: int = None  # type: ignore
    __text: Union[str, List[str], Tuple[str]] = None  # type: ignore
    __textvariable: tk.StringVar = None  # type: ignore
    __label_attr: Dict[str, Any] = None  # type: ignore

    def __init__(
        self,
        widget: tk.Misc,
        text: Union[str, List[str], Tuple[str], tk.StringVar] = "widget info",
        wait_time: int = 500,
        wrap_length: int = 0,
        **kwargs,
    ) -> None:
        """Create class object."""
        # set default attributes
        self.__label_attr = {
            "justify": tk.LEFT,
            "bg": "white",
            "relief": tk.SOLID,
            "borderwidth": 1,
        }
        # update attributes
        if kwargs:
            self.__label_attr.update(kwargs)

        self.__waittime = wait_time
        self.__wraplength = wrap_length
        self.__widget = widget

        # set message
        self.text = text
        self.__widget.bind("<Enter>", self.__enter)
        self.__widget.bind("<Leave>", self.__leave)
        self.__widget.bind("<ButtonPress>", self.__leave)

    def __enter(self, event: Optional[tk.Event] = None) -> None:
        """Call on <Enter> event."""
        self.__schedule()

    def __leave(self, event: Optional[tk.Event] = None) -> None:
        """Call on <Leave> event."""
        self.__unschedule()
        self.__hidetip()

    def __schedule(self) -> None:
        """Schedule method."""
        self.__unschedule()
        self.__id = self.__widget.after(self.__waittime, self.__showtip)

    def __unschedule(self) -> None:
        """Unschedule method."""
        __id: Optional[str] = self.__id
        self.__id = None
        if __id:
            self.__widget.after_cancel(__id)

    def __showtip(self, event: Optional[tk.Event] = None) -> None:
        """Show tooltip."""
        __x: int = 0
        __y: int = 0
        __cx: int
        __cy: int
        __x, __y, __cx, __cy = self.__widget.bbox("insert")  # type: ignore
        __x += self.__widget.winfo_rootx() + 25
        __y += self.__widget.winfo_rooty() + 20
        # creates a toplevel window
        self.__tw = tk.Toplevel(self.__widget)
        # Leaves only the label and removes the app window
        self.__tw.wm_overrideredirect(True)
        self.__tw.wm_geometry(f"+{__x}+{__y}")
        label = tk.Label(
            self.__tw,
            wraplength=self.__wraplength,
        )
        for key in self.__label_attr.keys():
            label[key.lower()] = self.__label_attr[key]
        if isinstance(self.text, tk.StringVar):
            label["textvariable"] = self.text
        else:
            label["text"] = self.text
        label.pack(ipadx=1)

    def __hidetip(self) -> None:
        """Hide tooltip."""
        __tw: Optional[Toplevel] = self.__tw
        self.__tw = None
        if __tw:
            __tw.destroy()

    @property
    def text(self) -> Union[str, tk.StringVar]:
        """Return text message."""
        if self.__text is None and self.__textvariable is None:
            self.__text = ""
        if self.__textvariable is None:
            if isinstance(self.__text, (List, Tuple)):
                tmp: str = ""
                for msg in self.__text:
                    tmp += msg if not tmp else f"\n{msg}"
                return tmp
            return self.__text
        else:
            return self.__textvariable

    @text.setter
    def text(self, value: Union[str, List[str], Tuple[str], tk.StringVar]) -> None:
        """Set text message object."""
        if isinstance(value, tk.StringVar):
            self.__textvariable = value
        else:
            self.__text = value


class VerticalScrolledTkFrame(tk.Frame, TkBase):
    """A pure Tkinter scrollable frame that actually works!
    * Use the 'interior' property to place widgets inside the scrollable frame.
    * Construct and pack/place/grid normally.
    * This frame only allows vertical scrolling.
    """

    __vscrollbar: tk.Scrollbar = None  # type: ignore
    __canvas: tk.Canvas = None  # type: ignore
    __interior: tk.Frame = None  # type: ignore
    __interior_id: int = None  # type: ignore

    def __init__(self, parent: tk.Misc, *args, **kw) -> None:
        tk.Frame.__init__(self, parent, *args, **kw)

        # Create a canvas object and a vertical scrollbar for scrolling it.
        # vscrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        self.__vscrollbar = tk.Scrollbar(self, orient=tk.VERTICAL)
        self.__vscrollbar.pack(fill=tk.Y, side=tk.RIGHT)
        self.__canvas = tk.Canvas(
            self, bd=0, highlightthickness=0, yscrollcommand=self.__vscrollbar.set
        )
        self.__canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.TRUE)
        self.__vscrollbar.config(command=self.__canvas.yview)

        # Reset the view
        self.__canvas.xview_moveto(0)
        self.__canvas.yview_moveto(0)

        # Create a frame inside the canvas which will be scrolled with it.
        # self.interior = interior = ttk.Frame(canvas)
        self.__interior = tk.Frame(self.__canvas)
        self.__interior_id: int = self.__canvas.create_window(
            0, 0, window=self.__interior, anchor=tk.NW
        )

        # Configure Events
        self.__interior.bind("<Configure>", self.__configure_interior)
        self.__canvas.bind("<Configure>", self.__configure_canvas)
        self.__canvas.bind("<Enter>", self.__bind_mouse)
        self.__canvas.bind("<Leave>", self.__unbind_mouse)

    @property
    def interior(self) -> tk.Frame:
        """The interior property."""
        return self.__interior

    def __configure_interior(self, event) -> None:
        # Update the scrollbars to match the size of the inner frame.
        self.__canvas.config(
            scrollregion=(
                0,
                0,
                self.__interior.winfo_reqwidth(),
                self.__interior.winfo_reqheight(),
            )
        )
        if self.__interior.winfo_reqwidth() != self.__canvas.winfo_width():
            # Update the canvas's width to fit the inner frame.
            self.__canvas.config(width=self.__interior.winfo_reqwidth())

    def __configure_canvas(self, event: tk.Event) -> None:
        # print(f"{event}")
        # print(f"{type(event)}")
        if self.__interior.winfo_reqwidth() != self.__canvas.winfo_width():
            # Update the inner frame's width to fill the canvas.
            self.__canvas.itemconfigure(
                self.__interior_id, width=self.__canvas.winfo_width()
            )

    def __bind_mouse(self, event: Optional[tk.Event] = None) -> None:
        # print(f"{event}")
        # print(f"{type(event)}")
        self.__canvas.bind_all("<4>", self.__on_mousewheel)
        self.__canvas.bind_all("<5>", self.__on_mousewheel)
        self.__canvas.bind_all("<MouseWheel>", self.__on_mousewheel)

    def __unbind_mouse(self, event: Optional[tk.Event] = None) -> None:
        # print(f"{event}")
        # print(f"{type(event)}")
        self.__canvas.unbind_all("<4>")
        self.__canvas.unbind_all("<5>")
        self.__canvas.unbind_all("<MouseWheel>")

    def __on_mousewheel(self, event: tk.Event) -> None:
        """Linux uses event.num; Windows / Mac uses event.delta"""
        # print(f"{event}")
        # print(f"{type(event)}")
        if event.num == 4 or event.delta > 0:
            self.__canvas.yview_scroll(-1, "units")
        elif event.num == 5 or event.delta < 0:
            self.__canvas.yview_scroll(1, "units")


class VerticalScrolledTtkFrame(ttk.Frame, TkBase):
    """A pure Tkinter scrollable frame that actually works!
    * Use the 'interior' property to place widgets inside the scrollable frame.
    * Construct and pack/place/grid normally.
    * This frame only allows vertical scrolling.
    """

    __vscrollbar: ttk.Scrollbar = None  # type: ignore
    __canvas: tk.Canvas = None  # type: ignore
    __interior: ttk.Frame = None  # type: ignore
    __interior_id: int = None  # type: ignore

    def __init__(self, parent: tk.Misc, *args, **kw) -> None:
        ttk.Frame.__init__(self, parent, *args, **kw)

        # Create a canvas object and a vertical scrollbar for scrolling it.
        # vscrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        self.__vscrollbar = ttk.Scrollbar(self, orient=tk.VERTICAL)
        self.__vscrollbar.pack(fill=tk.Y, side=tk.RIGHT)
        self.__canvas = tk.Canvas(
            self, bd=0, highlightthickness=0, yscrollcommand=self.__vscrollbar.set
        )
        self.__canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=tk.TRUE)
        self.__vscrollbar.config(command=self.__canvas.yview)

        # Reset the view
        self.__canvas.xview_moveto(0)
        self.__canvas.yview_moveto(0)

        # Create a frame inside the canvas which will be scrolled with it.
        # self.interior = interior = ttk.Frame(canvas)
        self.__interior = ttk.Frame(self.__canvas)
        self.__interior_id: int = self.__canvas.create_window(
            0, 0, window=self.__interior, anchor=tk.NW
        )

        # Configure Events
        self.__interior.bind("<Configure>", self.__configure_interior)
        self.__canvas.bind("<Configure>", self.__configure_canvas)
        self.__canvas.bind("<Enter>", self.__bind_mouse)
        self.__canvas.bind("<Leave>", self.__unbind_mouse)

    @property
    def interior(self) -> ttk.Frame:
        """The interior property."""
        return self.__interior

    def __configure_interior(self, event) -> None:
        # Update the scrollbars to match the size of the inner frame.
        self.__canvas.config(
            scrollregion=(
                0,
                0,
                self.__interior.winfo_reqwidth(),
                self.__interior.winfo_reqheight(),
            )
        )
        if self.__interior.winfo_reqwidth() != self.__canvas.winfo_width():
            # Update the canvas's width to fit the inner frame.
            self.__canvas.config(width=self.__interior.winfo_reqwidth())

    def __configure_canvas(self, event: tk.Event) -> None:
        # print(f"{event}")
        # print(f"{type(event)}")
        if self.__interior.winfo_reqwidth() != self.__canvas.winfo_width():
            # Update the inner frame's width to fill the canvas.
            self.__canvas.itemconfigure(
                self.__interior_id, width=self.__canvas.winfo_width()
            )

    def __bind_mouse(self, event: Optional[tk.Event] = None) -> None:
        # print(f"{event}")
        # print(f"{type(event)}")
        self.__canvas.bind_all("<4>", self.__on_mousewheel)
        self.__canvas.bind_all("<5>", self.__on_mousewheel)
        self.__canvas.bind_all("<MouseWheel>", self.__on_mousewheel)

    def __unbind_mouse(self, event: Optional[tk.Event] = None) -> None:
        # print(f"{event}")
        # print(f"{type(event)}")
        self.__canvas.unbind_all("<4>")
        self.__canvas.unbind_all("<5>")
        self.__canvas.unbind_all("<MouseWheel>")

    def __on_mousewheel(self, event: tk.Event) -> None:
        """Linux uses event.num; Windows / Mac uses event.delta"""
        # print(f"{event}")
        # print(f"{type(event)}")
        if event.num == 4 or event.delta > 0:
            self.__canvas.yview_scroll(-1, "units")
        elif event.num == 5 or event.delta < 0:
            self.__canvas.yview_scroll(1, "units")


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 04.09.2023

  Purpose: logs subsystem classes.
"""

import time
import threading
from inspect import currentframe

from typing import Optional, Tuple

from jsktoolbox.attribtool import NoDynamicAttributes, ReadOnlyClass
from jsktoolbox.raisetool import Raise
from jsktoolbox.libs.base_logs import (
    BLoggerQueue,
    LoggerQueue,
    LogsLevelKeys,
    Keys,
)
from jsktoolbox.libs.base_th import ThBaseObject
from jsktoolbox.logstool.engines import *


class _Keys(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    LCO = "__LCO__"
    LEO = "__LEO__"


class LoggerClient(BLoggerQueue, NoDynamicAttributes):
    """Logger Client main class."""

    def __init__(
        self, queue: Optional[LoggerQueue] = None, name: Optional[str] = None
    ) -> None:
        """Constructor.

        Arguments:
        queue [LoggerQueue] - optional LoggerQeueu class object from LoggerEngine, required, but can be set after the object is created,
        name [str] - optional client name for logs decorator
        """
        # store name
        self.name = name
        # logger queue
        if queue:
            self.logs_queue = queue

    @property
    def name(self) -> Optional[str]:
        """Get LoggerClient name string."""
        if Keys.NAME not in self._data:
            self._data[Keys.NAME] = None
        return self._data[Keys.NAME]

    @name.setter
    def name(self, name: Optional[str]) -> None:
        """Set LoggerClient name string."""
        if name and not isinstance(name, str):
            raise Raise.error(
                f"Expected 'name' as string type, received: '{type(name)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[Keys.NAME] = name

    def message(self, message: str, log_level: str = LogsLevelKeys.INFO) -> None:
        """Send message to logging subsystem."""
        if not isinstance(log_level, str):
            raise Raise.error(
                f"Expected 'log_level' as string type, received: '{type(log_level)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        if log_level not in LogsLevelKeys.keys:
            raise Raise.error(
                f"Expected 'log_level' as key from .base_logs.LogsLevelKeys.keys, received: '{log_level}'.",
                KeyError,
                self._c_name,
                currentframe(),
            )
        if not isinstance(message, str):
            message = f"{message}"
        if self.name is not None:
            message = f"[{self.name}] {message}"
        if self.logs_queue:
            self.logs_queue.put(message, log_level)

    @property
    def message_alert(self) -> None:
        """Do nothing, for defining setter only."""

    @message_alert.setter
    def message_alert(self, message: str) -> None:
        """Property for sending messages with ALERT level."""
        self.message(message, LogsLevelKeys.ALERT)

    @property
    def message_critical(self) -> None:
        """Do nothing, for defining setter only."""

    @message_critical.setter
    def message_critical(self, message: str) -> None:
        """Property for sending messages with CRITICAL level."""
        self.message(message, LogsLevelKeys.CRITICAL)

    @property
    def message_debug(self) -> None:
        """Do nothing, for defining setter only."""

    @message_debug.setter
    def message_debug(self, message: str) -> None:
        """Property for sending messages with DEBUG level."""
        self.message(message, LogsLevelKeys.DEBUG)

    @property
    def message_emergency(self) -> None:
        """Do nothing, for defining setter only."""

    @message_emergency.setter
    def message_emergency(self, message: str) -> None:
        """Property for sending messages with EMERGENCY level."""
        self.message(message, LogsLevelKeys.EMERGENCY)

    @property
    def message_error(self) -> None:
        """Do nothing, for defining setter only."""

    @message_error.setter
    def message_error(self, message: str) -> None:
        """Property for sending messages with ERROR level."""
        self.message(message, LogsLevelKeys.ERROR)

    @property
    def message_info(self) -> None:
        """Do nothing, for defining setter only."""

    @message_info.setter
    def message_info(self, message: str) -> None:
        """Property for sending messages with INFO level."""
        self.message(message, LogsLevelKeys.INFO)

    @property
    def message_notice(self) -> None:
        """Do nothing, for defining setter only."""

    @message_notice.setter
    def message_notice(self, message: str) -> None:
        """Property for sending messages with NOTICE level."""
        self.message(message, LogsLevelKeys.NOTICE)

    @property
    def message_warning(self) -> None:
        """Do nothing, for defining setter only."""

    @message_warning.setter
    def message_warning(self, message: str) -> None:
        """Property for sending messages with WARNING level."""
        self.message(message, LogsLevelKeys.WARNING)


class LoggerEngine(BLoggerQueue, NoDynamicAttributes):
    """LoggerEngine container class."""

    def __init__(self) -> None:
        """Constructor."""
        # make logs queue object
        self.logs_queue = LoggerQueue()
        # default logs level configuration
        self._data[Keys.NO_CONF] = {}
        self._data[Keys.NO_CONF][LogsLevelKeys.INFO] = [LoggerEngineStdout()]
        self._data[Keys.NO_CONF][LogsLevelKeys.WARNING] = [LoggerEngineStdout()]
        self._data[Keys.NO_CONF][LogsLevelKeys.NOTICE] = [LoggerEngineStdout()]
        self._data[Keys.NO_CONF][LogsLevelKeys.DEBUG] = [LoggerEngineStderr()]
        self._data[Keys.NO_CONF][LogsLevelKeys.ERROR] = [
            LoggerEngineStdout(),
            LoggerEngineStderr(),
        ]
        self._data[Keys.NO_CONF][LogsLevelKeys.CRITICAL] = [
            LoggerEngineStdout(),
            LoggerEngineStderr(),
        ]

    def add_engine(self, log_level: str, engine: ILoggerEngine) -> None:
        """Add LoggerEngine to specific log level.

        Arguments:
        log_level [str] - String Key from .base_log.LogsLevelKeys.keys,
        engine [ILoggerEngine] - an object created from Engine classes.
        """
        if not isinstance(log_level, str):
            raise Raise.error(
                f"Expected 'log_level' as string type, received: '{type(log_level)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        if not isinstance(engine, ILoggerEngine):
            raise Raise.error(
                f"Expected ILoggerEngine type, received: '{type(engine)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        if Keys.CONF not in self._data:
            self._data[Keys.CONF] = {}
            self._data[Keys.CONF][log_level] = [engine]
        else:
            if log_level not in self._data[Keys.CONF].keys():
                self._data[Keys.CONF][log_level] = [engine]
            else:
                test = False
                for i in range(0, len(self._data[Keys.CONF][log_level])):
                    if (
                        self._data[Keys.CONF][log_level][i].__class__
                        == engine.__class__
                    ):
                        self._data[Keys.CONF][log_level][i] = engine
                        test = True
                if not test:
                    self._data[Keys.CONF][log_level].append(engine)

    def send(self) -> None:
        """The LoggerEngine method.

        For sending messages to the configured logging subsystem.
        """
        while True:
            if self.logs_queue is not None:
                item: Optional[tuple[str, ...]] = self.logs_queue.get()
                if item is not None:
                    # get tuple(log_level, message)
                    log_level, message = item
                    # check if has have configured logging subsystem
                    if Keys.CONF in self._data and len(self._data[Keys.CONF]) > 0:
                        if log_level in self._data[Keys.CONF]:
                            for leng in self._data[Keys.CONF][log_level]:
                                engine: ILoggerEngine = leng
                                engine.send(message)
                    else:
                        if log_level in self._data[Keys.NO_CONF]:
                            for leng in self._data[Keys.NO_CONF][log_level]:
                                engine: ILoggerEngine = leng
                                engine.send(message)
                else:
                    return None
            else:
                return None


class ThLoggerProcessor(threading.Thread, ThBaseObject, NoDynamicAttributes):
    """LoggerProcessor thread class."""

    def __init__(self, debug: bool = False) -> None:
        """Constructor."""
        threading.Thread.__init__(self, name=self._c_name)
        self._stop_event = threading.Event()
        self._debug = debug
        self.daemon = True
        self.sleep_period = 0.2

    @property
    def logger_engine(self) -> Optional[LoggerEngine]:
        """Return LoggerEngine object if any."""
        if _Keys.LEO not in self._data:
            self._data[_Keys.LEO] = None
        return self._data[_Keys.LEO]

    @logger_engine.setter
    def logger_engine(self, obj: LoggerEngine) -> None:
        """Set LoggerEngine object."""
        if not isinstance(obj, LoggerEngine):
            raise Raise.error(
                f"Excpected LoggerEngine type, received: '{type(obj)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.LEO] = obj
        if self.logger_client and self.logger_engine and self.logger_engine.logs_queue:
            self.logger_client.logs_queue = self.logger_engine.logs_queue

    @property
    def logger_client(self) -> Optional[LoggerClient]:
        """Return LoggerClient object if any."""
        if _Keys.LCO not in self._data:
            self._data[_Keys.LCO] = None
        return self._data[_Keys.LCO]

    @logger_client.setter
    def logger_client(self, obj: LoggerClient) -> None:
        """Set LoggerEngine object."""
        if not isinstance(obj, LoggerClient):
            raise Raise.error(
                f"Expected LoggerClient type, received: '{type(obj)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.LCO] = obj
        if (
            self.logger_engine
            and self.logger_engine.logs_queue
            and self.logger_client
            and obj.logs_queue is None
        ):
            self.logger_client.logs_queue = self.logger_engine.logs_queue

    def run(self) -> None:
        """Start the procedure."""
        # check list
        if self.logger_engine is None:
            raise Raise.error(
                "LoggerEngine not set.",
                ValueError,
                self._c_name,
                currentframe(),
            )
        if self.logger_client is None:
            raise Raise.error(
                "LoggerClient not set.",
                ValueError,
                self._c_name,
                currentframe(),
            )
        if self._debug:
            self.logger_client.message_debug = f"[{self._c_name}] Start."
        # run
        while not self.stopped:
            self.logger_engine.send()
            time.sleep(self.sleep_period)
        if self._debug:
            self.logger_client.message_debug = f"[{self._c_name}] Stop."
        self.logger_engine.send()

    def stop(self) -> None:
        """Set stop event."""
        if self._debug and self.logger_client:
            self.logger_client.message_debug = f"[{self._c_name}] stopping..."
        if self._stop_event:
            self._stop_event.set()

    @property
    def stopped(self) -> bool:
        """Return stop flag."""
        if self._stop_event:
            return self._stop_event.is_set()
        return True


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 25.08.2023

  Purpose: Connector interfaces module.
"""

from abc import ABC, abstractmethod
from typing import List, Union


class IConnect(ABC):
    """Conection class interface."""

    @abstractmethod
    def connect(self) -> bool:
        """Connection method."""

    @abstractmethod
    def disconnect(self) -> bool:
        """Disconnect method."""

    @abstractmethod
    def errors(self) -> List:
        """Get list or errors after executed commands."""

    @abstractmethod
    def execute(self, commands: Union[str, List]) -> bool:
        """Execute method."""

    @property
    @abstractmethod
    def is_alive(self) -> bool:
        """Get alive flag from connected protocol."""

    @property
    @abstractmethod
    def login(self) -> str:
        """Get login property."""

    @login.setter
    @abstractmethod
    def login(self, username: str) -> None:
        """Set login property."""

    @abstractmethod
    def outputs(self) -> List:
        """Get list of results after executed commands."""

    @property
    @abstractmethod
    def password(self) -> str:
        """Get password property."""

    @password.setter
    @abstractmethod
    def password(self, passwordstring: str) -> None:
        """Set password property."""

    @property
    @abstractmethod
    def prototype(self) -> str:
        """Get protocol type property."""


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
  Created: 10.10.2023

  Purpose: LogFormatter classes.
"""

import time
from datetime import datetime

from jsktoolbox.libs.base_logs import BLogFormatter
from jsktoolbox.datetool import Timestamp


class LogFormatterNull(BLogFormatter):
    """Log Formatter Null class."""

    def __init__(self) -> None:
        """Constructor."""
        self._forms_.append("{message}")
        self._forms_.append("[{name}]: {message}")


class LogFormatterDateTime(BLogFormatter):
    """Log Formatter DateTime class."""

    def __init__(self) -> None:
        """Constructor."""
        self._forms_.append(self.__get_formated_date__)
        self._forms_.append("{message}")
        self._forms_.append("[{name}]: {message}")

    def __get_formated_date__(self) -> str:
        return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


class LogFormatterTime(BLogFormatter):
    """Log Formatter Time class."""

    def __init__(self) -> None:
        """Constructor."""
        self._forms_.append(self.__get_formated_time__)
        self._forms_.append("{message}")
        self._forms_.append("[{name}]: {message}")

    def __get_formated_time__(self) -> str:
        return datetime.now().strftime("%H:%M:%S")


class LogFormatterTimestamp(BLogFormatter):
    """Log Formatter Timestamp class."""

    def __init__(self) -> None:
        """Constructor."""
        self._forms_.append(Timestamp.now)
        self._forms_.append("{message}")
        self._forms_.append("[{name}]: {message}")


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 05.09.2023

  Purpose: various classes of interaction with the system.
"""

import os
import sys
import getopt

from inspect import currentframe
from pathlib import Path
from typing import Optional, Union, List, Tuple, Dict

from jsktoolbox.attribtool import NoDynamicAttributes, ReadOnlyClass
from jsktoolbox.raisetool import Raise
from jsktoolbox.libs.base_data import BData


class _Keys(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    ARGS = "__args__"
    CONFIGURED_ARGS = "__cargs__"
    DESC_OPTS = "__desc_opts__"
    EXAMPLE_OPTS = "__ex_opts__"
    EXISTS = "__exists__"
    ISDIR = "__is_dir__"
    ISFILE = "__is_file__"
    ISSYMLINK = "__is_symlink__"
    LIST = "__list__"
    LONG_OPTS = "__long_opts__"
    PATHNAME = "__pathname__"
    POSIXPATH = "__posix_path__"
    SHORT_OPTS = "__short_opts__"
    SPLIT = "__split__"


class CommandLineParser(BData, NoDynamicAttributes):
    """Parser for command line options."""

    def __init__(self) -> None:
        """Constructor."""
        self._data[_Keys.CONFIGURED_ARGS] = {}
        self._data[_Keys.ARGS] = {}

    def configure_argument(
        self,
        short_arg: str,
        long_arg: str,
        desc_arg: Optional[Union[str, List, Tuple]] = None,
        has_value: bool = False,
        example_value: Optional[str] = None,
    ) -> None:
        """Application command line argument configuration method and its description."""
        if _Keys.SHORT_OPTS not in self._data[_Keys.CONFIGURED_ARGS]:
            self._data[_Keys.CONFIGURED_ARGS][_Keys.SHORT_OPTS] = ""
        if _Keys.LONG_OPTS not in self._data[_Keys.CONFIGURED_ARGS]:
            self._data[_Keys.CONFIGURED_ARGS][_Keys.LONG_OPTS] = []
        if _Keys.DESC_OPTS not in self._data[_Keys.CONFIGURED_ARGS]:
            self._data[_Keys.CONFIGURED_ARGS][_Keys.DESC_OPTS] = []
        if _Keys.EXAMPLE_OPTS not in self._data[_Keys.CONFIGURED_ARGS]:
            self._data[_Keys.CONFIGURED_ARGS][_Keys.EXAMPLE_OPTS] = []

        if not short_arg:
            short_arg = "_"

        if not long_arg:
            raise Raise.error(
                f"Long argument name is required.",
                AttributeError,
                self._c_name,
                currentframe(),
            )

        self._data[_Keys.CONFIGURED_ARGS][_Keys.SHORT_OPTS] += short_arg + (
            ":" if has_value else ""
        )
        self._data[_Keys.CONFIGURED_ARGS][_Keys.LONG_OPTS].append(
            long_arg + ("=" if has_value else "")
        )

        tmp = ""
        if desc_arg:
            if isinstance(desc_arg, str):
                tmp = desc_arg
            elif isinstance(desc_arg, (Tuple, List)):
                tmp = []
                for desc in desc_arg:
                    tmp.append(desc)
                if not tmp:
                    tmp = ""
            else:
                tmp = str(desc_arg)
        self._data[_Keys.CONFIGURED_ARGS][_Keys.DESC_OPTS].append(tmp)

        tmp = ""
        if example_value:
            if isinstance(example_value, str):
                tmp = example_value

        self._data[_Keys.CONFIGURED_ARGS][_Keys.EXAMPLE_OPTS].append(tmp)

    def parse_arguments(self) -> bool:
        """Command line arguments parser."""
        # replace ':' if exists in short option string
        short_mod = str(self._data[_Keys.CONFIGURED_ARGS][_Keys.SHORT_OPTS]).replace(
            ":", ""
        )
        long_mod = []
        for item in self._data[_Keys.CONFIGURED_ARGS][_Keys.LONG_OPTS]:
            long_mod.append(item.replace("=", ""))

        try:
            opts, _ = getopt.getopt(
                sys.argv[1:],
                self._data[_Keys.CONFIGURED_ARGS][_Keys.SHORT_OPTS],
                self._data[_Keys.CONFIGURED_ARGS][_Keys.LONG_OPTS],
            )
        except getopt.GetoptError as ex:
            print(f"Command line argument error: {ex}")
            return False

        for opt, value in opts:
            for short_arg, long_arg in zip(
                short_mod,
                long_mod,
            ):
                if opt in ("-" + short_arg, "--" + long_arg):
                    self.args[long_arg] = value
        return True

    def get_option(self, option: str) -> Optional[str]:
        """Get value of the option or None if it doesn't exist."""
        out = self.args.get(option)
        if out is None:
            return None
        return str(self.args.get(option))

    def dump(self) -> Dict:
        """Dump configured data structure as Dict:
        {'long opt name':{'short':str, 'has_value':bool, 'description':str, 'example':str}}
        """
        out = {}
        short_mod = str(self._data[_Keys.CONFIGURED_ARGS][_Keys.SHORT_OPTS]).replace(
            ":", ""
        )

        for short_arg, long_arg, desc_arg, ex_arg in zip(
            short_mod,
            self._data[_Keys.CONFIGURED_ARGS][_Keys.LONG_OPTS],
            self._data[_Keys.CONFIGURED_ARGS][_Keys.DESC_OPTS],
            self._data[_Keys.CONFIGURED_ARGS][_Keys.EXAMPLE_OPTS],
        ):
            out[long_arg] = {
                "short": short_arg if short_arg != "_" else "",
                "has_value": True if long_arg[-1] == "=" else False,
                "description": desc_arg,
                "example": ex_arg,
            }
        return out

    @property
    def args(self) -> Dict:
        """Return parsed arguments dict."""
        return self._data[_Keys.ARGS]


class Env(NoDynamicAttributes):
    """Environment class."""

    @classmethod
    @property
    def home(cls) -> str:
        """Return home dir name."""
        tmp: Optional[str] = os.getenv("HOME")
        if tmp:
            return tmp
        return ""

    @classmethod
    @property
    def username(cls) -> str:
        """Return login name."""
        tmp: Optional[str] = os.getenv("USER")
        if tmp:
            return tmp
        return ""


class PathChecker(BData, NoDynamicAttributes):
    """PathChecker class for filesystem path."""

    def __init__(self, pathname: str, check_deep: bool = True) -> None:
        """Constructor."""
        if pathname is None:
            raise Raise.error(
                "Expected 'pathname' as string, not None.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        if not isinstance(pathname, str):
            raise Raise.error(
                f"Expected 'pathname' as string, received: '{type(pathname)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        if isinstance(pathname, str) and len(pathname) == 0:
            raise Raise.error(
                "'pathname' cannot be an empty string.",
                ValueError,
                self._c_name,
                currentframe(),
            )
        self._data[_Keys.PATHNAME] = pathname
        self._data[_Keys.SPLIT] = check_deep
        self._data[_Keys.LIST] = []
        # analysis
        self.__run__()

    def __run__(self) -> None:
        """Path analysis procedure."""
        query = Path(self._data[_Keys.PATHNAME])
        # check exists
        self._data[_Keys.EXISTS] = query.exists()
        if self._data[_Keys.EXISTS]:
            # check isfile
            self._data[_Keys.ISFILE] = query.is_file()
            # check isdir
            self._data[_Keys.ISDIR] = query.is_dir()
            # check issymlink
            self._data[_Keys.ISSYMLINK] = query.is_symlink()
            # resolve symlink
            self._data[_Keys.POSIXPATH] = str(query.resolve())

        if self._data[_Keys.SPLIT]:
            # split and analyse
            tmp: str = ""
            for item in self.path.split(os.sep):
                if item == "":
                    continue
                tmp += f"{os.sep}{item}"
                self._data[_Keys.LIST].append(PathChecker(tmp, False))

    def __str__(self) -> str:
        """Return class data as string."""
        return (
            "PathChecker("
            f"'pathname': '{self.path}', "
            f"'exists': '{self.exists}', "
            f"'is_dir': '{self.is_dir if self.exists else ''}', "
            f"'is_file': '{self.is_file if self.exists else ''}', "
            f"'is_symlink': '{self.is_symlink if self.exists else ''}', "
            f"'posixpath': '{self.posixpath if self.exists else ''}'"
            ")"
        )

    def __repr__(self) -> str:
        """Return string representation."""
        return f"PathChecker('{self.path}')"

    @property
    def dirname(self) -> Optional[str]:
        """Return dirname from path."""
        if self.exists:
            last: Optional[str] = None
            for item in self._data[_Keys.LIST]:
                if item.is_dir:
                    last = item.path
            return last
        return None

    @property
    def filename(self) -> Optional[str]:
        """Return filename from path."""
        if self.exists and self.is_file:
            tmp: list[str] = self.path.split(os.sep)
            if len(tmp) > 0:
                if tmp[-1] != "":
                    return tmp[-1]
        return None

    @property
    def exists(self) -> bool:
        """Return path exists flag."""
        if _Keys.EXISTS in self._data:
            return self._data[_Keys.EXISTS]
        else:
            raise Raise.error(
                "Unexpected exception",
                KeyError,
                self._c_name,
                currentframe(),
            )

    @property
    def is_dir(self) -> bool:
        """Return path isdir flag."""
        if _Keys.ISDIR in self._data:
            return self._data[_Keys.ISDIR]
        else:
            raise Raise.error(
                "Unexpected exception",
                KeyError,
                self._c_name,
                currentframe(),
            )

    @property
    def is_file(self) -> bool:
        """Return path isfile flag."""
        if _Keys.ISFILE in self._data:
            return self._data[_Keys.ISFILE]
        else:
            raise Raise.error(
                "Unexpected exception",
                KeyError,
                self._c_name,
                currentframe(),
            )

    @property
    def is_symlink(self) -> bool:
        """Return path issymlink flag."""
        if _Keys.ISSYMLINK in self._data:
            return self._data[_Keys.ISSYMLINK]
        else:
            raise Raise.error(
                "Unexpected exception",
                KeyError,   
                self._c_name,
                currentframe(),
            )

    @property
    def path(self) -> str:
        """Return path string."""
        return self._data[_Keys.PATHNAME]

    @property
    def posixpath(self) -> Optional[str]:
        """Return path string."""
        if self.exists:
            return self._data[_Keys.POSIXPATH]
        return None

    def create(self) -> bool:
        """Create path procedure."""
        test_path: str = self.path
        file = True
        if self.path[-1] == os.sep:
            file = False
            test_path = self.path[:-1]
        for item in self._data[_Keys.LIST]:
            if item.exists:
                continue
            if item.path == test_path:
                # last element
                if file:
                    # touch file
                    with open(item.path, "w"):
                        pass
                else:
                    os.mkdir(item.path)
            else:
                os.mkdir(item.path)
        # check
        self._data[_Keys.LIST] = []
        self.__run__()

        return self.exists


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 04.09.2023

  Purpose: base class for log subsystem.
"""

import syslog

from inspect import currentframe
from typing import Optional, Tuple, List, Dict, Any

from jsktoolbox.libs.base_data import BData, BClasses
from jsktoolbox.attribtool import NoDynamicAttributes, ReadOnlyClass
from jsktoolbox.raisetool import Raise


class Keys(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    BUFFERED = "__buffered__"
    CONF = "__conf__"
    DIR = "__dir__"
    FACILITY = "__facility__"
    FILE = "__file__"
    FORMATTER = "__formatter__"
    LEVEL = "__level__"
    NAME = "__name__"
    NO_CONF = "__noconf__"
    QUEUE = "__queue__"
    SYSLOG = "__syslog__"


class SysLogKeys(object, metaclass=ReadOnlyClass):
    """SysLog keys definition container class."""

    class __Levels(object, metaclass=ReadOnlyClass):
        ALERT = syslog.LOG_ALERT
        CRITICAL = syslog.LOG_CRIT
        DEBUG = syslog.LOG_DEBUG
        EMERGENCY = syslog.LOG_EMERG
        ERROR = syslog.LOG_ERR
        INFO = syslog.LOG_INFO
        NOTICE = syslog.LOG_NOTICE
        WARNING = syslog.LOG_WARNING

    class __Facilities(object, metaclass=ReadOnlyClass):
        DAEMON = syslog.LOG_DAEMON
        LOCAL0 = syslog.LOG_LOCAL0
        LOCAL1 = syslog.LOG_LOCAL1
        LOCAL2 = syslog.LOG_LOCAL2
        LOCAL3 = syslog.LOG_LOCAL3
        LOCAL4 = syslog.LOG_LOCAL4
        LOCAL5 = syslog.LOG_LOCAL5
        LOCAL6 = syslog.LOG_LOCAL6
        LOCAL7 = syslog.LOG_LOCAL7
        MAIL = syslog.LOG_MAIL
        SYSLOG = syslog.LOG_SYSLOG
        USER = syslog.LOG_USER

    @classmethod
    @property
    def level(cls) -> type[__Levels]:
        """Returns Levels keys property."""
        return cls.__Levels

    @classmethod
    @property
    def facility(cls) -> type[__Facilities]:
        """Returns Facility keys property."""
        return cls.__Facilities

    @classmethod
    @property
    def level_keys(cls) -> Dict:
        """Returns level keys property."""
        return {
            "ALERT": SysLogKeys.level.ALERT,
            "CRITICAL": SysLogKeys.level.CRITICAL,
            "DEBUG": SysLogKeys.level.DEBUG,
            "EMERGENCY": SysLogKeys.level.EMERGENCY,
            "ERROR": SysLogKeys.level.ERROR,
            "INFO": SysLogKeys.level.INFO,
            "NOTICE": SysLogKeys.level.NOTICE,
            "WARNING": SysLogKeys.level.WARNING,
        }

    @classmethod
    @property
    def facility_keys(cls) -> Dict:
        """Returns Facility keys property."""
        return {
            "DAEMON": SysLogKeys.facility.DAEMON,
            "LOCAL0": SysLogKeys.facility.LOCAL0,
            "LOCAL1": SysLogKeys.facility.LOCAL1,
            "LOCAL2": SysLogKeys.facility.LOCAL2,
            "LOCAL3": SysLogKeys.facility.LOCAL3,
            "LOCAL4": SysLogKeys.facility.LOCAL4,
            "LOCAL5": SysLogKeys.facility.LOCAL5,
            "LOCAL6": SysLogKeys.facility.LOCAL6,
            "LOCAL7": SysLogKeys.facility.LOCAL7,
            "MAIL": SysLogKeys.facility.MAIL,
            "SYSLOG": SysLogKeys.facility.SYSLOG,
            "USER": SysLogKeys.facility.USER,
        }


class LogsLevelKeys(object, metaclass=ReadOnlyClass):
    """LogsLevelKeys container class."""

    ALERT = "ALERT"
    CRITICAL = "CRITICAL"
    DEBUG = "DEBUG"
    EMERGENCY = "EMERGENCY"
    ERROR = "ERROR"
    INFO = "INFO"
    NOTICE = "NOTICE"
    WARNING = "WARNING"

    @classmethod
    @property
    def keys(cls) -> tuple:
        """Return tuple of avaiable keys."""
        return tuple(
            [
                LogsLevelKeys.ALERT,
                LogsLevelKeys.CRITICAL,
                LogsLevelKeys.DEBUG,
                LogsLevelKeys.EMERGENCY,
                LogsLevelKeys.ERROR,
                LogsLevelKeys.INFO,
                LogsLevelKeys.NOTICE,
                LogsLevelKeys.WARNING,
            ]
        )


class LoggerQueue(BClasses, NoDynamicAttributes):
    """LoggerQueue simple class."""

    __queue: List[List[str]] = []

    def __init__(self) -> None:
        """Constructor."""
        self.__queue = []

    def get(self) -> Optional[tuple[str, ...]]:
        """Get item from queue.

        Returs queue tuple[log_level:str, message:str] or None if empty.
        """
        try:
            return tuple(self.__queue.pop(0))
        except IndexError:
            return None
        except Exception as ex:
            raise Raise.error(
                f"Unexpected exception was thrown: {ex}",
                Exception,
                self._c_name,
                currentframe(),
            )

    def put(self, message: str, log_level: str = LogsLevelKeys.INFO) -> None:
        """Put item to queue."""
        if log_level not in LogsLevelKeys.keys:
            raise Raise.error(
                f"logs_level key not found, '{log_level}' received.",
                KeyError,
                self._c_name,
                currentframe(),
            )
        self.__queue.append(
            [
                log_level,
                message,
            ]
        )


class BLoggerQueue(BData, NoDynamicAttributes):
    """Logger Queue base metaclass."""

    @property
    def logs_queue(self) -> Optional[LoggerQueue]:
        """Get LoggerQueue object."""
        if Keys.QUEUE not in self._data:
            return None
        return self._data[Keys.QUEUE]

    @logs_queue.setter
    def logs_queue(self, obj: LoggerQueue) -> None:
        """Set LoggerQueue object."""
        if not isinstance(obj, LoggerQueue):
            raise Raise.error(
                f"Expected LoggerQueue type, received: '{type(obj)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        self._data[Keys.QUEUE] = obj


class BLoggerEngine(BData, NoDynamicAttributes):
    """Base class for LoggerEngine classes."""

    @property
    def name(self) -> Optional[str]:
        """Return app name string."""
        if Keys.NAME not in self._data:
            self._data[Keys.NAME] = None
        return self._data[Keys.NAME]

    @name.setter
    def name(self, value: str) -> None:
        """Set app name string."""
        self._data[Keys.NAME] = value


class BLogFormatter(NoDynamicAttributes):
    """Log formatter base class."""

    __template: Optional[str] = None
    __forms: Optional[List] = None

    def format(self, message: str, name: Optional[str] = None) -> str:
        """Method for format message string.

        Arguments:
        message [str]: log string to send
        name [str]: optional name of apps,
        """
        out: str = ""
        for item in self._forms_:
            if callable(item):
                out += f"{item()} "
            elif isinstance(item, str):
                if name is None:
                    if item.find("name") == -1:
                        out += item.format(message=f"{message}")
                else:
                    if item.find("name") > 0:
                        out += item.format(
                            name=f"{name}",
                            message=f"{message}",
                        )
        return out

    @property
    def _forms_(self) -> List:
        """Get forms list."""
        if self.__forms is None:
            self.__forms = []
        return self.__forms

    @_forms_.setter
    def _forms_(self, item: Any) -> None:
        """Set forms list."""
        self._forms_.append(item)


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 02.07.2023

  Purpose: Base classes for restricting the creation of dynamic attributes
  on instance of derived types.

  [NoNewAttributes]
  The solution idea published in: Python Cookbook (2004), A. Martelli,
  A. Ravenscroft, D. Ascher

  [NoDynamicAttributes]
  Another solution with the same functionality.
"""

from typing import Any, Callable


def _no_new_attributes(
    wrapped_setattr: Any,
) -> Callable[[Any, str, Any], None]:
    """Internal function for use in the current module only."""

    def __setattr__(self, name: str, value: Any) -> None:
        """Check if the attribute is defined, throw an exception if not."""
        if hasattr(self, name):
            wrapped_setattr(self, name, value)
        else:
            raise AttributeError(
                f"Undefined attribute {name} cannot be added to {self}"
            )

    return __setattr__


class NoNewAttributes:
    """NoNewAttributes - base class.

    Class for restricting the creation of dynamic attributes on instances
    of derived types.
    """

    __setattr__: Callable[[Any, str, Any], None] = _no_new_attributes(
        object.__setattr__
    )

    class __metaclass__(type):
        __setattr__: Callable[[Any, str, Any], None] = _no_new_attributes(
            type.__setattr__
        )


class NoDynamicAttributes:
    """NoDynamicAttributes - base class.

    Class for restricting the creation of dynamic attributes on instances
    of derived types.
    """

    def __setattr__(self, name: str, value: Any) -> None:
        if not hasattr(self, name):
            raise AttributeError(
                f"Cannot add new attribute '{name}' to {self.__class__.__name__} object"
            )
        super().__setattr__(name, value)


class ReadOnlyClass(type):
    """ReadOnlyClass - metaclass for creating read only classes.

    example:
    class A(object, metaclass=ReadOnlyClass):
          foo = "don't change me"
    """

    def __setattr__(self, name: str, value: Any) -> None:
        raise ValueError(f"Read only attribute: {name}.")


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
  Created: 02.12.2023

  Purpose: Sets of classes for various date/time operations.
"""

from time import time
from datetime import datetime, timezone, timedelta
from typing import Optional, Union
from inspect import currentframe

from jsktoolbox.attribtool import NoNewAttributes
from jsktoolbox.raisetool import Raise


class DateTime(NoNewAttributes):
    """DateTime class for generating various datetime structures."""

    @classmethod
    def now(cls, tz: Optional[timezone] = None) -> datetime:
        """Return datetime.datetime.now() object.

        Argument:
        tz [datetime.timezone] - datetime.timezone.utc for UTC, default None for current set timezone.
        """
        return datetime.now(tz=tz)

    @classmethod
    def datetime_from_timestamp(
        cls,
        timestamp_seconds: Union[int, float],
        tz: Optional[timezone] = None,
    ) -> datetime:
        """Returns datetime from timestamp int."""
        if not isinstance(timestamp_seconds, (int, float)):
            raise Raise.error(
                f"Expected int or float type, received: '{type(timestamp_seconds)}'.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        return datetime.fromtimestamp(timestamp_seconds, tz=tz)

    @classmethod
    def elapsed_time_from_seconds(cls, seconds: Union[int, float]) -> timedelta:
        """Convert given seconds to timedelta structure."""
        if not isinstance(seconds, (int, float)):
            raise Raise.error(
                f"Expected int or float type, received: '{type(seconds)}'.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        return timedelta(seconds=seconds)

    @classmethod
    def elapsed_time_from_timestamp(
        cls, seconds: Union[int, float], tz: Optional[timezone] = None
    ) -> timedelta:
        """Generate date/time timedelta with elapsed time, from given timestamp to now.

        WARNING:
        Returns the timedelta accurate to the second.
        """
        if not isinstance(seconds, (int, float)):
            raise Raise.error(
                f"Expected int or float type, received: '{type(seconds)}'.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        out: timedelta = cls.now(tz=tz) - datetime.fromtimestamp(seconds, tz=tz)
        return timedelta(days=out.days, seconds=out.seconds)


class Timestamp(NoNewAttributes):
    """Timestamp class for geting current timestamp."""

    @classmethod
    @property
    def now(cls) -> int:
        """Return timestamp int."""
        return int(time())


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
  Created: 29.10.2023

  Purpose: Main class for creating and processes config files.
"""

import re
from inspect import currentframe
from typing import List, Dict, Optional, Any
from jsktoolbox.attribtool import NoDynamicAttributes, ReadOnlyClass
from jsktoolbox.raisetool import Raise
from jsktoolbox.libs.base_data import BData
from jsktoolbox.configtool.libs.file import FileProcessor
from jsktoolbox.configtool.libs.data import DataProcessor
from jsktoolbox.configtool.libs.data import SectionModel


class _Keys(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    DESC = "__desc__"
    DP = "__data_processor__"
    FP = "__file_processor__"
    RE_BOOL = "__re_bool__"
    RE_DESC = "__re_description__"
    RE_FALSE = "__re_false__"
    RE_FLOAT = "__re_float__"
    RE_INT = "__re_integer__"
    RE_LIST = "__re_list__"
    RE_SECTION = "__re_section__"
    RE_TRUE = "__re_true__"
    RE_VAR = "__re_variable__"
    VALUE = "__value__"
    VARNAME = "__varname__"


class Config(BData, NoDynamicAttributes):
    """Config main class."""

    def __init__(
        self,
        filename: str,
        main_section_name: str,
        auto_create: bool = False,
    ) -> None:
        """Constructor."""
        self._data[_Keys.FP] = FileProcessor()
        self._data[_Keys.DP] = DataProcessor()
        self.__fp.file = filename
        self.__dp.main_section = main_section_name
        if auto_create:
            if not self.__fp.file_exists:
                self.__fp.file_create()
        # compile regex
        self._data[_Keys.RE_SECTION] = re.compile(r"\s{0,}\[.*\]\s{0,}")
        self._data[_Keys.RE_DESC] = re.compile(r"\s{0,}#")
        self._data[_Keys.RE_VAR] = re.compile(r"\s{0,}\S{1,}\s{0,}=")
        self._data[_Keys.RE_INT] = re.compile(r"^\d{1,}$")
        self._data[_Keys.RE_FLOAT] = re.compile(r"^\d{1,}\.\d{1,}$")
        self._data[_Keys.RE_BOOL] = re.compile(r"^true|false|yes|no$", re.IGNORECASE)
        self._data[_Keys.RE_TRUE] = re.compile(r"^true|yes$", re.IGNORECASE)
        self._data[_Keys.RE_FALSE] = re.compile(r"^false|no$", re.IGNORECASE)
        self._data[_Keys.RE_LIST] = re.compile(r"^\[.*\]$")

    @property
    def __fp(self) -> FileProcessor:
        """Return FileProcessor object."""
        return self._data[_Keys.FP]

    @property
    def __dp(self) -> DataProcessor:
        """Return DataProcessor object."""
        return self._data[_Keys.DP]

    @property
    def file_exists(self) -> bool:
        """Check if file exists."""
        return self.__fp.file_exists

    def __value_parser(self, item: str) -> Any:
        """Return proper type of value."""
        if self._data[_Keys.RE_BOOL].match(item):
            return True if self._data[_Keys.RE_TRUE].match(item) else False
        elif self._data[_Keys.RE_INT].match(item):
            return int(item)
        elif self._data[_Keys.RE_FLOAT].match(item):
            return float(item)
        elif self._data[_Keys.RE_LIST].match(item):
            out = []
            tmp: list[str] = [x.strip() for x in item.strip("[]").split(",")]
            for item in tmp:
                out.append(self.__value_parser(item))
            return out
        return str(item.strip("\"'"))

    def __var_parser(self, line: str) -> Dict:
        """Return Dict[varname, value, desc]."""
        out: dict[str, Any] = {
            _Keys.VARNAME: None,
            _Keys.VALUE: None,
            _Keys.DESC: None,
        }
        tmp: list[str] = line.split("=", 1)
        if len(tmp) != 2:
            raise Raise.error(
                f"Unexpected config line format: '{line}'",
                ValueError,
                self._c_name,
                currentframe(),
            )
        out[_Keys.VARNAME] = tmp[0].strip()
        if len(tmp[1]) > 0:
            tmp = tmp[1].split("#", 1)
            # desc
            if len(tmp) == 2 and len(tmp[1]) > 0:
                out[_Keys.DESC] = tmp[1].strip()
            # value
            out[_Keys.VALUE] = self.__value_parser(tmp[0].strip())

        return out

    def load(self) -> bool:
        """Load config file to DataProcessor."""
        test = False
        # 1. load file into list
        file: List[str] = self.__fp.readlines()
        section_name: str = ""
        if self.__dp.main_section is not None:
            section_name = self.__dp.main_section
        for line in file:
            # check section
            if self._data[_Keys.RE_SECTION].match(line):
                section_name = self.__dp.add_section(line)
            # check description
            elif self._data[_Keys.RE_DESC].match(line):
                self.__dp.set(section_name, desc=line.strip("# "))
            # check var
            elif self._data[_Keys.RE_VAR].match(line):
                out = self.__var_parser(line)
                self.__dp.set(
                    section=section_name,
                    varname=out[_Keys.VARNAME],
                    value=out[_Keys.VALUE],
                    desc=out[_Keys.DESC],
                )
            else:
                self.__dp.set(section_name, desc=line)
            test = True
        return test

    def save(self) -> bool:
        """Save config file from DataProcessor."""
        test = False
        self.__fp.write(self.__dp.dump)
        test = True
        return test

    def get(
        self, section: str, varname: Optional[str] = None, desc: bool = False
    ) -> Any:
        """Get and return data."""
        return self.__dp.get(section, varname, desc)

    def set(
        self,
        section: str,
        varname: Optional[str] = None,
        value: Optional[Any] = None,
        desc: Optional[str] = None,
    ) -> None:
        """Set data."""
        self.__dp.set(section, varname, value, desc)

    def has_section(self, section: str) -> bool:
        """Check section name in config file."""
        if not isinstance(section, str):
            raise Raise.error(
                f"Expected String type, received: '{type(section)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return self.__dp.get_section(section) is not None

    def has_varname(self, section_name: str, varname: str) -> bool:
        """Check varname in section."""
        if not isinstance(varname, str):
            raise Raise.error(
                f"Expected String type, received: '{type(varname)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        if self.has_section(section_name):
            tmp: Optional[SectionModel] = self.__dp.get_section(section_name)
            if tmp is not None:
                found_section: SectionModel = tmp
                return found_section.get_variable(varname) is not None
        return False

    @property
    def main_section_name(self) -> Optional[str]:
        """Return main section name string."""
        return self.__dp.main_section


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 23.06.2023

  Purpose: Octet class for representing ipv4 octet.
"""
from inspect import currentframe
from typing import Union, TypeVar
from jsktoolbox.attribtool import NoDynamicAttributes
from jsktoolbox.raisetool import Raise

from jsktoolbox.libs.interfaces.comparators import IComparators
from jsktoolbox.libs.base_data import BClasses

TOctet = TypeVar("TOctet", bound="Octet")


class Octet(IComparators, BClasses, NoDynamicAttributes):
    """Class for representing ipv4 octet.

    Constructor arguments:
    value [str|int|Octet] -- Value of octet in proper range from 0 to 255

    Public property:
    value [int] -- Return integer representation of octet.

    Public setters:
    value [str|int|Octet] -- Set value of octet.
    """

    __value: int = 0

    def __init__(self, value: Union[str, int, TOctet]) -> None:
        """Constructor."""
        self.value = value

    def __eq__(self, arg: TOctet) -> bool:
        """Equal."""
        return int(self) == int(arg)

    def __ge__(self, arg: TOctet) -> bool:
        """Greater then or equal."""
        return int(self) >= int(arg)

    def __gt__(self, arg: TOctet) -> bool:
        """Greater then."""
        return int(self) > int(arg)

    def __le__(self, arg: TOctet) -> bool:
        """Less then or equal."""
        return int(self) <= int(arg)

    def __lt__(self, arg: TOctet) -> bool:
        """Less then."""
        return int(self) < int(arg)

    def __ne__(self, arg: TOctet) -> bool:
        """Negative."""
        return int(self) != int(arg)

    def __int__(self) -> int:
        """Return integer representation of octet."""
        return self.value

    def __str__(self) -> str:
        """Return string representation of octet."""
        return str(self.value)

    def __repr__(self) -> str:
        """Return representation of object."""
        return f"{self._c_name}({self.value})"

    @staticmethod
    def __check_range(value: int) -> bool:
        if value not in range(0, 256):
            return False
        return True

    @staticmethod
    def __is_integer(value: str) -> bool:
        try:
            int(value)
            return True
        except:
            return False

    @property
    def value(self) -> int:
        """Rerutn value of Octet as int."""
        return self.__value

    @value.setter
    def value(
        self,
        args: Union[str, int, TOctet],
    ) -> None:
        if isinstance(args, int):
            if Octet.__check_range(args):
                self.__value = args
                return
            else:
                raise Raise.error(
                    f"Received value '{args}' out of range(0-255).",
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
        elif isinstance(args, str):
            if Octet.__is_integer(args):
                var = int(args)
                if Octet.__check_range(var):
                    self.__value = var
                    return
                else:
                    raise Raise.error(
                        f"Received value '{args}' out of range(0-255).",
                        ValueError,
                        self._c_name,
                        currentframe(),
                    )
        elif isinstance(args, Octet):
            tmp: TOctet = args
            self.__value = tmp.value
            return
        raise Raise.error(
            f"Expected Integer or String type, received: '{type(args)}'.",
            TypeError,
            self._c_name,
            currentframe(),
        )


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
  Created: 14.09.2023

  Purpose: Classes for IPv6

  https://www.ibm.com/docs/en/ts3500-tape-library?topic=formats-subnet-masks-ipv4-prefixes-ipv6
"""

import socket

from copy import deepcopy
from inspect import currentframe
from typing import TypeVar, Union, List

from jsktoolbox.attribtool import NoDynamicAttributes
from jsktoolbox.raisetool import Raise
from .libs.words import Word16
from jsktoolbox.libs.interfaces.comparators import IComparators
from jsktoolbox.libs.base_data import BClasses

TAddress6 = TypeVar("TAddress6", bound="Address6")


class Address6(IComparators, BClasses, NoDynamicAttributes):
    """Address6 class for representing IPv6 adresses.

    Constructor arguments:
    addr: Union[str, int, List[Word16]] -- IPv6 address representation as string, integer or list of eight Word16

    Public property:
    words: List[Word16] -- Return list of eight Word16

    Public setter:
    words: Union[str, int, List] -- Set IPv6 address from string, integer or list of Word16.
    """

    __varint: int = 0

    def __init__(
        self, addr: Union[str, int, Union[List[int], List[str], List[Word16]]]
    ) -> None:
        """Constructor."""
        self.words = addr

    def __eq__(self, arg: TAddress6) -> bool:
        """Equal."""
        return int(self) == int(arg)

    def __ge__(self, arg: TAddress6) -> bool:
        """Greater or equal."""
        return int(self) >= int(arg)

    def __gt__(self, arg: TAddress6) -> bool:
        """Greater."""
        return int(self) > int(arg)

    def __le__(self, arg: TAddress6) -> bool:
        """Less or equal."""
        return int(self) <= int(arg)

    def __lt__(self, arg: TAddress6) -> bool:
        """Less."""
        return int(self) < int(arg)

    def __ne__(self, arg: TAddress6) -> bool:
        """Negative."""
        return int(self) != int(arg)

    @staticmethod
    def __check_groups(group_list: List[str]) -> List[str]:
        for i in range(0, len(group_list)):
            group_list[i] = group_list[i].zfill(4)
        return group_list

    @staticmethod
    def __expand_ipv6(ipv6_address: str) -> str:
        # Sprawdzenie czy adres jest już w pełnej reprezentacji
        if "::" not in ipv6_address:
            return ipv6_address

        # Podziel adres na dwie części, przed '::' i po '::'
        parts: list[str] = ipv6_address.split("::", 1)
        head: List[str] = Address6.__check_groups(parts[0].split(":"))
        tail: List[str] = Address6.__check_groups(parts[1].split(":"))

        # Oblicz ile zer należy dodać, aby osiągnąć 8 części
        missing_zeros: int = 8 - (len(head) + len(tail))
        zero_part = "0000"

        # Połącz części przed '::' i po '::', wstawiając brakujące zera
        expanded_parts: str = ":".join(head + [zero_part] * missing_zeros + tail)

        return expanded_parts

    @staticmethod
    def __is_valid_ipv6(ipv6_addr: str) -> bool:
        """Check if ipv6_addr is valid."""
        try:
            # Używamy socket.inet_pton, aby sprawdzić poprawność adresu IPv6
            socket.inet_pton(socket.AF_INET6, ipv6_addr)
            return True
        except (socket.error, ValueError):
            return False

    @staticmethod
    def __int_to_ip(ipint: int) -> str:
        """Convert ip int representation to ipv6 str."""
        # W przypadku adresu IPv6 zawsze przekształcamy go na 16 bajtów (128 bitów)
        binary_ip: bytes = ipint.to_bytes(16, byteorder="big")
        ipv6_address: str = socket.inet_ntop(socket.AF_INET6, binary_ip)

        return ipv6_address

    @staticmethod
    def __ip_to_int(ipstr: str) -> int:
        """Convert ipv6 str representation to ip int."""
        # Używamy socket.inet_pton, aby przekształcić adres IPv6 w postać binarną
        packed_ip: bytes = socket.inet_pton(socket.AF_INET6, ipstr)
        # Następnie przekształcamy binarny adres IPv6 na liczbę całkowitą (integer)
        int_ip: int = int.from_bytes(packed_ip, byteorder="big")

        return int_ip

    def __set_words_from_list(
        self, value: Union[List[int], List[str], List[Word16]]
    ) -> None:
        """Set address from list."""
        if len(value) != 8:
            raise Raise.error(
                f"Expected list of eight elements.",
                ValueError,
                self._c_name,
                currentframe(),
            )
        tmp: str = (
            f"{str(Word16(value[0]))}:"
            f"{str(Word16(value[1]))}:"
            f"{str(Word16(value[2]))}:"
            f"{str(Word16(value[3]))}:"
            f"{str(Word16(value[4]))}:"
            f"{str(Word16(value[5]))}:"
            f"{str(Word16(value[6]))}:"
            f"{str(Word16(value[7]))}"
        )
        self.__set_words_from_str(tmp)

    def __set_words_from_int(self, value: int) -> None:
        # if value >= 0 and value <= 340282366920938463463374607431768211455:
        if value in range(0, 340282366920938463463374607431768211456):
            self.__varint = value
        else:
            raise Raise.error(
                f"IP-int out of range (0-340282366920938463463374607431768211455), received: {value}",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __set_words_from_str(self, value: str) -> None:
        if Address6.__is_valid_ipv6(value):
            self.__varint = Address6.__ip_to_int(value)
        else:
            raise Raise.error(
                f"IPv6 address is invalid: {value}",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __int__(self) -> int:
        """Return ipv4 representation as integer."""
        return self.__varint

    def __str__(self) -> str:
        """Return string representation of address."""
        return Address6.__int_to_ip(self.__varint)

    def __repr__(self) -> str:
        """Return representation of object."""
        return f"{self._c_name}('{str(self)}')"

    @property
    def words(self) -> List[Word16]:
        """Return words list of eight Word16."""
        tmp: list[str] = Address6.__expand_ipv6(str(self)).split(":")
        return [
            Word16(tmp[0]),
            Word16(tmp[1]),
            Word16(tmp[2]),
            Word16(tmp[3]),
            Word16(tmp[4]),
            Word16(tmp[5]),
            Word16(tmp[6]),
            Word16(tmp[7]),
        ]

    @words.setter
    def words(
        self, value: Union[str, int, Union[List[int], List[str], List[Word16]]]
    ) -> None:
        if isinstance(value, List):
            self.__set_words_from_list(value)
        elif isinstance(value, int):
            self.__set_words_from_int(value)
        elif isinstance(value, str):
            self.__set_words_from_str(value)
        else:
            raise Raise.error(
                f"Expected String or Integer or List type, received: {type(value)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )


# prefix
# https://www.heficed.com/subnet-mask-cheat-sheet/

TPrefix6 = TypeVar("TPrefix6", bound="Prefix6")


class Prefix6(IComparators, BClasses, NoDynamicAttributes):
    """Prefix6 class for IPv6 addresses.

    Constructor argument:
    prefix: Union[str, int] -- Set prefix from string or integer.

    Public property:
    prefix: str -- Return prefix as string.

    Public setter:
    prefix: Union[str, int] -- Set prefix from string or integer.
    """

    __prefix_int: int = 0

    def __init__(self, prefix: Union[str, int]) -> None:
        """Constructor."""
        self.prefix = prefix

    def __eq__(self, arg: TPrefix6) -> bool:
        """Equal."""
        return int(self) == int(arg)

    def __ge__(self, arg: TPrefix6) -> bool:
        """Greater or equal."""
        return int(self) >= int(arg)

    def __gt__(self, arg: TPrefix6) -> bool:
        """Greater."""
        return int(self) > int(arg)

    def __le__(self, arg: TPrefix6) -> bool:
        """Less or equal."""
        return int(self) <= int(arg)

    def __lt__(self, arg: TPrefix6) -> bool:
        """Less."""
        return int(self) < int(arg)

    def __ne__(self, arg: TPrefix6) -> bool:
        """Negative."""
        return int(self) != int(arg)

    def __str__(self) -> str:
        """Return prefix as string."""
        return str(self.__prefix_int)

    def __int__(self) -> int:
        """Return prefix as integer."""
        return self.__prefix_int

    def __repr__(self) -> str:
        """Return Prefix6 representation string."""
        return f"{self._c_name}({int(self)})"

    def __range_validator(self, value: int) -> bool:
        """Proper range validator."""
        if value not in range(0, 129):
            raise Raise.error(
                f"Prefix out of range (0-128), received: {value}",
                ValueError,
                self._c_name,
                currentframe(),
            )
        return True

    @staticmethod
    def __is_integer(value: str) -> bool:
        try:
            int(value)
            return True
        except:
            return False

    @property
    def prefix(self) -> str:
        """Return prefix as string."""
        return str(self)

    @prefix.setter
    def prefix(self, value: Union[str, int]) -> None:
        """Set prefix from string or integer."""
        if isinstance(value, int) and self.__range_validator(value):
            self.__prefix_int = value
        elif isinstance(value, str):
            if Prefix6.__is_integer(value) and self.__range_validator(int(value)):
                self.__prefix_int = int(value)
            else:
                raise Raise.error(
                    f"Expected proper integer string, '{value}' received.",
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
        else:
            raise Raise.error(
                f"Expected String or Integer type, received: '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )


# Network
class Network6(BClasses, NoDynamicAttributes):
    """Network6 IPv6 class.

    Constructor argument:
    addr: Union[str, List] -- Set IPv6 network address from string or two element list of address [Address6,str,int,list] and prefix [Prefix6, str, int].

    Public property:
    address: Address6 -- Return IPv6 address set in the constructor.
    count: int -- Return count hosts adresses in network range.
    hosts: List[Address6] -- Return hosts address list.
    network: Address6 -- Return network address.
    prefix: Prefix6 -- Return prefix.
    max: Address6 -- Return max address of host in network range.
    min: Address6 -- Return min address of host in network range.
    """

    __address: Address6 = None  # type: ignore
    __prefix: Prefix6 = None  # type: ignore

    def __init__(self, addr: Union[str, List]) -> None:
        """Constructor."""
        if isinstance(addr, str):
            self.__network_from_str(addr)
        elif isinstance(addr, List):
            self.__network_from_list(addr)
        else:
            raise Raise.error(
                f"Expected IP network string or list type, received: '{type(addr)}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __str__(self) -> str:
        """Return string representation of network address."""
        return f"{self.network}/{int(self.prefix)}"

    def __repr__(self) -> str:
        """Return  string representation of class object."""
        return f"{self._c_name}('{str(self)}')"

    def __network_from_str(self, addr: str) -> None:
        """Build configuration from string."""
        if addr.find("/") > 0:
            tmp: list[str] = addr.split("/")
            self.__address = Address6(tmp[0])
            self.__prefix = Prefix6(tmp[1])
        else:
            raise Raise.error(
                f"Expected network address in 'ipv6/prefix' format string, received: '{addr}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __network_from_list(self, addr: List) -> None:
        """Build configuration from list."""
        if len(addr) != 2:
            raise Raise.error(
                "A list of two elements was expected: ['ipv6','prefix']",
                ValueError,
                self._c_name,
                currentframe(),
            )
        if isinstance(addr[0], Address6):
            self.__address = deepcopy(addr[0])
        else:
            self.__address = Address6(addr[0])
        if isinstance(addr[1], Prefix6):
            self.__prefix = deepcopy(addr[1])
        else:
            self.__prefix = Prefix6(addr[1])

    @property
    def address(self) -> Address6:
        """Return IPv6 address."""
        return self.__address

    @property
    def count(self) -> int:
        """Return number of hosts in subnet."""
        return 2 ** (128 - int(self.prefix))

    @property
    def hosts(self) -> List[Address6]:
        """Return list of hosts addresses."""
        start = int(self.min)
        tmp: List[Address6] = []
        for i in range(0, self.count):
            tmp.append(Address6(start + i))
        return tmp

    @property
    def max(self) -> Address6:
        """Return last IPv6 address from subnet."""
        return Address6(int(self.network) + self.count - 1)

    @property
    def min(self) -> Address6:
        """Return first IPv6 address from subnet."""
        ip = int(self.address)
        mask: int = (1 << 128 - int(self.prefix)) - 1
        net: int = ip & ~mask
        subnet_address = []
        for _ in range(8):
            subnet_address.insert(0, format(net & 0xFFFF, "x"))
            net >>= 16
        return Address6(":".join(subnet_address))

    @property
    def network(self) -> Address6:
        """Return network address (first host address of subnet)."""
        return self.min

    @property
    def prefix(self) -> Prefix6:
        """Return IPv6 network prefix."""
        return self.__prefix


# SubNetwork
class SubNetwork6(BClasses, NoDynamicAttributes):
    """SubNetwork6 calculator class.

    Constructor argument:
    network: Network6 -- The address of the network where the subnet is being searched for.
    prefix: Prefix6 -- Subnet prefix.

    Public property:
    subnets: List[Network6] -- Subnet list.
    """

    __network: Network6 = None  # type: ignore
    __prefix: Prefix6 = None  # type: ignore

    def __init__(self, network: Network6, prefix: Prefix6) -> None:
        """Constructor."""
        if isinstance(network, Network6) and isinstance(prefix, Prefix6):
            if int(network.prefix) <= int(prefix):
                self.__network = network
                self.__prefix = prefix
            else:
                raise Raise.error(
                    (
                        "The network prefix must be greater then or equal to the subnet prefix you are looking for."
                        f"Received: {int(network.prefix)} and {int(prefix)}"
                    ),
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
        else:
            raise Raise.error(
                f"Argument of (Network6, Prefix6) expected, ({type(network)},{type(prefix)}) received.",
                TypeError,
                self._c_name,
                currentframe(),
            )

    @property
    def subnets(self) -> List[Network6]:
        """Return subnets list."""
        tmp: List[Network6] = []
        nstart = int(self.__network.min)
        nend = int(self.__network.max)
        start: int = nstart
        while True:
            subnet = Network6([Address6(start), self.__prefix])
            tmp.append(subnet)
            if int(subnet.max) >= nend:
                break
            start = int(subnet.max) + 1
        return tmp


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 23.06.2023

  Purpose: Classes for IPv4
"""

import socket
import struct
from copy import deepcopy
from inspect import currentframe
from typing import TypeVar, Union, List, Optional

from jsktoolbox.attribtool import NoDynamicAttributes
from jsktoolbox.raisetool import Raise
from .libs.octets import Octet
from jsktoolbox.libs.interfaces.comparators import IComparators
from jsktoolbox.libs.base_data import BClasses

TAddress = TypeVar("TAddress", bound="Address")


class Address(IComparators, BClasses, NoDynamicAttributes):
    """Address class for representing IPv4 adresses.

    Constructor arguments:
    addr: Union[str, int, List[Octets]] -- IPv4 address representation as string, integer or list of four Octets

    Public property:
    octets: List[Octet] -- Return list of four Octets

    Public setter:
    octets: Union[str, int, List] -- Set IPv4 address from string, integer or list of octets.
    """

    __varint: int = 0

    def __init__(
        self, addr: Union[str, int, Union[List[str], List[int], List[Octet]]]
    ) -> None:
        """Constructor."""
        self.octets = addr

    def __eq__(self, arg: TAddress) -> bool:
        """Equal."""
        return int(self) == int(arg)

    def __ge__(self, arg: TAddress) -> bool:
        """Greater or equal."""
        return int(self) >= int(arg)

    def __gt__(self, arg: TAddress) -> bool:
        """Greater."""
        return int(self) > int(arg)

    def __le__(self, arg: TAddress) -> bool:
        """Less or equal."""
        return int(self) <= int(arg)

    def __lt__(self, arg: TAddress) -> bool:
        """Less."""
        return int(self) < int(arg)

    def __ne__(self, arg: TAddress) -> bool:
        """Negative."""
        return int(self) != int(arg)

    @staticmethod
    def __int_to_ip(ipint: int) -> str:
        """Convert ip int representation to ipv4 str."""
        return socket.inet_ntoa(struct.pack("!L", ipint))

    @staticmethod
    def __ip_to_int(ipstr: str) -> int:
        """Convert ipv4 str representation to ip int."""
        return struct.unpack("!L", socket.inet_aton(ipstr))[0]

    def __set_octets_from_list(
        self, value: Union[List[str], List[int], List[Octet]]
    ) -> None:
        if not value:
            raise Raise.error(
                "Empty list received.",
                ValueError,
                self._c_name,
                currentframe(),
            )
        if len(value) != 4:
            raise Raise.error(
                f"Expected list with four elements, len({len(value)}) received.",
                ValueError,
                self._c_name,
                currentframe(),
            )

        self.__varint = Address.__ip_to_int(
            f"{Octet(value[0])}.{Octet(value[1])}.{Octet(value[2])}.{Octet(value[3])}"
        )

    def __set_octets_from_int(self, value: int) -> None:
        if value in range(0, 4294967296):
            self.__varint = value
        else:
            raise Raise.error(
                f"IP-int out of range (0-4294967295), received: {value}",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __set_octets_from_str(self, value: str) -> None:
        self.__set_octets_from_list(value.split("."))

    def __int__(self) -> int:
        """Return ipv4 representation as integer."""
        return self.__varint

    def __str__(self) -> str:
        """Return string representation of address."""
        return Address.__int_to_ip(self.__varint)

    def __repr__(self):
        """Return representation of object."""
        return f"{self._c_name}('{str(self)}')"

    @property
    def octets(self) -> List[Octet]:
        """Return octets list of four Octets."""
        tmp: list[str] = str(self).split(".")
        return [Octet(tmp[0]), Octet(tmp[1]), Octet(tmp[2]), Octet(tmp[3])]

    @octets.setter
    def octets(
        self, value: Union[str, int, Union[List[str], List[int], List[Octet]]]
    ) -> None:
        if isinstance(value, List):
            self.__set_octets_from_list(value)
        elif isinstance(value, int):
            self.__set_octets_from_int(value)
        elif isinstance(value, str):
            self.__set_octets_from_str(value)
        else:
            raise Raise.error(
                f"Expected String or Integer or List type, received: {type(value)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )


# netmask
class Netmask(BClasses, NoDynamicAttributes):
    """Netmask class for IPv4 addresses.

    Constructor argument:
    addr: Union[str, int, List] -- Set netmask from string, integer or list of proper format of netmask octets.

    Public property:
    octets: List[Octet] -- Return netmask as list of four octets.
    cidr: str -- Return netmask in CIDR string format.

    Public setter:
    octets: List[Octet] -- Set netmask from list of 4 values [int||str||Octets].
    cidr: Union[str, int] -- Set netmask from CIDR format of string or integer.
    """

    # CIDR format
    __cidr: int = 0

    def __init__(
        self, addr: Union[str, int, Union[List[str], List[int], List[Octet]]]
    ) -> None:
        """Constructor."""
        if isinstance(addr, int):
            self.cidr = addr
        elif isinstance(addr, str):
            if len(addr) < 3:
                self.cidr = addr
            else:
                self.octets = addr
        elif isinstance(addr, List):
            self.octets = str(Address(addr))
        else:
            raise Raise.error(
                f"Expected String, Integer or List type, received: '{type(addr)}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __int__(self) -> int:
        return self.__cidr

    def __str__(self) -> str:
        # convert CIDR to netmask
        return socket.inet_ntoa(
            struct.pack("!I", (1 << 32) - (1 << (32 - self.__cidr)))
        )

    def __repr__(self) -> str:
        return f"{self._c_name}({self.cidr})"

    def __cidr_validator(self, cidr: int) -> None:
        """Check and set cidr."""
        if cidr >= 0 and cidr <= 32:
            self.__cidr = cidr
        else:
            raise Raise.error(
                f"CIDR is out of range (0-32), received: {cidr}",
                ValueError,
                self._c_name,
                currentframe(),
            )

    @staticmethod
    def __octets_validator(octets: int) -> bool:
        """Check if given octets list is valid."""
        test: list[int] = [
            0,  # '0.0.0.0'
            2147483648,  # "128.0.0.0",
            3221225472,  # "192.0.0.0",
            3758096384,  # "224.0.0.0",
            4026531840,  # "240.0.0.0",
            4160749568,  # "248.0.0.0",
            4227858432,  # "252.0.0.0",
            4261412864,  # "254.0.0.0",
            4278190080,  # "255.0.0.0",
            4286578688,  # "255.128.0.0",
            4290772992,  # "255.192.0.0",
            4292870144,  # "255.224.0.0",
            4293918720,  # "255.240.0.0",
            4294443008,  # "255.248.0.0",
            4294705152,  # "255.252.0.0",
            4294836224,  # "255.254.0.0",
            4294901760,  # "255.255.0.0",
            4294934528,  # "255.255.128.0",
            4294950912,  # "255.255.192.0",
            4294959104,  # "255.255.224.0",
            4294963200,  # "255.255.240.0",
            4294965248,  # "255.255.248.0",
            4294966272,  # "255.255.252.0",
            4294966784,  # "255.255.254.0",
            4294967040,  # "255.255.255.0",
            4294967168,  # "255.255.255.128",
            4294967232,  # "255.255.255.192",
            4294967264,  # "255.255.255.224",
            4294967280,  # "255.255.255.240",
            4294967288,  # "255.255.255.248",
            4294967292,  # "255.255.255.252",
            4294967294,  # "255.255.255.254",
            4294967295,  # "255.255.255.255",
        ]
        if octets in test:
            return True
        return False

    @property
    def octets(self) -> List[Octet]:
        """Return octets list of four Octets."""
        tmp: list[str] = str(self).split(".")
        return [Octet(tmp[0]), Octet(tmp[1]), Octet(tmp[2]), Octet(tmp[3])]

    @octets.setter
    def octets(
        self, addr: Union[str, int, Union[List[str], List[int], List[Octet]]]
    ) -> None:
        """Set netmask from list of 4 values [int||str||Octets]."""
        tmp = int(Address(addr))
        if not Netmask.__octets_validator(tmp):
            raise Raise.error(
                f"Invalid mask, received: {str(Address(addr))}",
                ValueError,
                self._c_name,
                currentframe(),
            )
        self.cidr = sum([bin(x.value).count("1") for x in Address(addr).octets])

    @property
    def cidr(self) -> str:
        """Return CIDR netmask as string type."""
        return str(self.__cidr)

    @cidr.setter
    def cidr(self, value: Union[str, int]) -> None:
        if isinstance(value, str) and value.isdigit():
            self.__cidr_validator(int(value))
        elif isinstance(value, int):
            self.__cidr_validator(value)
        else:
            raise Raise.error(
                f"Expected Digit string or Int, received: {value}",
                ValueError,
                self._c_name,
                currentframe(),
            )


# Network
class Network(BClasses, NoDynamicAttributes):
    """Network IPv4 class.

    Constructor argument:
    addr: Union[str, List] -- Set IPv4 network address from string or two element list of address [Address,str,int,list] and netmask [Netmask, str, int, list].

    Public property:
    address: Address -- Return IPv4 address set in the constructor.
    broadcast: Address -- Return broadcast address.
    count: int -- Return count hosts adresses in network range.
    hosts: List[Address] -- Return hosts address list.
    network: Address -- Return network address.
    mask: Netmask -- Return netmask.
    max: Address -- Return max address of host in network range.
    min: Address -- Return min address of host in network range.
    """

    __address: Address = None  # type: ignore
    __mask: Netmask = None  # type: ignore

    def __init__(self, addr: Union[str, List]) -> None:
        """Constructor."""
        if isinstance(addr, str):
            self.__network_from_str(addr)
        elif isinstance(addr, List):
            self.__network_from_list(addr)
        else:
            raise Raise.error(
                f"Expected IP network string or list type, received: '{type(addr)}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __str__(self) -> str:
        """Return string representation of network address."""
        return f"{self.network}/{int(self.mask)}"

    def __repr__(self) -> str:
        """Return  string representation of class object."""
        return f"{self._c_name}({str(self)})"

    def __network_from_str(self, addr: str) -> None:
        """Build configuration from string."""
        if addr.find("/") > 0:
            tmp: list[str] = addr.split("/")
            self.__address = Address(tmp[0])
            self.__mask = Netmask(tmp[1])
        else:
            raise Raise.error(
                f"Expected network address in 'ip/mask' format string, received: '{addr}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __network_from_list(self, addr: List) -> None:
        """Build configuration from list."""
        if len(addr) != 2:
            raise Raise.error(
                "Expected two element list ['ip','netmask']",
                ValueError,
                self._c_name,
                currentframe(),
            )
        if isinstance(addr[0], Address):
            self.__address = deepcopy(addr[0])
        else:
            self.__address = Address(addr[0])
        if isinstance(addr[1], Netmask):
            self.__mask = deepcopy(addr[1])
        else:
            self.__mask = Netmask(addr[1])

    @property
    def address(self) -> Address:
        """Return IPv4 address."""
        return self.__address

    @property
    def broadcast(self) -> Address:
        """Return broadcast address."""
        ip = int(self.address)
        mask = int(Address(self.mask.octets))
        broadcast: int = ip | (mask ^ (1 << 32) - 1)
        return Address(broadcast)

    @property
    def count(self) -> int:
        """Return count hosts adresses in network range."""
        net = int(self.network)
        broadcast = int(self.broadcast)
        return broadcast - net - 1 if broadcast - net > 2 else 0

    @property
    def hosts(self) -> List[Address]:
        """Return list of hosts in network range."""
        tmp: List[Address] = []
        net = int(self.network)
        broadcast = int(self.broadcast)
        for i in range(1, broadcast - net):
            tmp.append(Address(net + i))
        return tmp

    @property
    def mask(self) -> Netmask:
        """Return IPv4 network mask."""
        return self.__mask

    @property
    def max(self) -> Address:
        """Return last address of host in network range."""
        net = int(self.network)
        broadcast = int(self.broadcast)
        ip = broadcast - 1
        return Address(ip) if ip > net else self.broadcast

    @property
    def min(self) -> Address:
        """Return first host address in network range."""
        net = int(self.network)
        broadcast = int(self.broadcast)
        ip: int = net + 1
        return Address(ip) if ip < broadcast else self.network

    @property
    def network(self) -> Address:
        """Return network address."""
        ip = int(self.address)
        mask = int(Address(self.mask.octets))
        net: int = ip & mask
        return Address(net)


# SubNetwork
class SubNetwork(BClasses, NoDynamicAttributes):
    """SubNetwork calculator class.

    Constructor argument:
    network: Network -- The address of the network where the subnet is being searched for.
    mask: Netmask -- Subnet mask.

    Public property:
    subnets: List[Network] -- Subnet list.
    """

    __network: Network = None  # type: ignore
    __mask: Netmask = None  # type: ignore

    def __init__(self, network: Network, mask: Netmask) -> None:
        """Constructor."""
        if isinstance(network, Network) and isinstance(mask, Netmask):
            if int(network.mask) <= int(mask):
                self.__network = network
                self.__mask = mask
            else:
                raise Raise.error(
                    (
                        "The network mask must be greater then or equal to the subnet mask you are looking for."
                        f"Received: {int(network.mask)} and {int(mask)}"
                    ),
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
        else:
            raise Raise.error(
                f"Expected argument of (Network, Netmask), received: ({type(network)},{type(mask)}).",
                TypeError,
                self._c_name,
                currentframe(),
            )

    @property
    def subnets(self) -> List[Network]:
        """Return subnets list."""
        tmp: List[Network] = []
        nstart = int(self.__network.network)
        nend = int(self.__network.broadcast)
        start: int = nstart
        while True:
            subnet = Network([Address(start), self.__mask])
            tmp.append(subnet)
            if int(subnet.broadcast) >= nend:
                break
            start = int(subnet.broadcast) + 1
        return tmp


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
  Created: 04.11.2023

  Purpose: Logger Engine interface class.
"""

from abc import ABC, abstractmethod


class ILoggerEngine(ABC):
    """Logger engine interface class."""

    @abstractmethod
    def send(self, message: str) -> None:
        """Send message method."""


# #[EOF]#######################################################################



# -*- coding: utf-8 -*-
"""
  layout.py
  Author : Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
  Created: 15.01.2024, 13:48:01
  
  Purpose: Tkinter geometry managers keys helper classes.
"""

import tkinter as tk

from jsktoolbox.attribtool import ReadOnlyClass


class Pack(object, metaclass=ReadOnlyClass):
    """Pack geometry manager.

    https://www.pythontutorial.net/tkinter/tkinter-pack/
    """

    class Anchor(object, metaclass=ReadOnlyClass):
        """The anchor parameter allows you to anchor the widget to the edge of the allocated space."""

        CENTER = tk.CENTER
        E = tk.E
        N = tk.N
        NE = tk.NE
        NW = tk.NW
        S = tk.S
        SE = tk.SE
        SW = tk.SW
        W = tk.W

    class Side(object, metaclass=ReadOnlyClass):
        """The side parameter determines the direction of the widgets in the pack layout."""

        BOTTOM = tk.BOTTOM
        LEFT = tk.LEFT
        RIGHT = tk.RIGHT
        TOP = tk.TOP

    class Fill(object, metaclass=ReadOnlyClass):
        """The fill determines if a widget will occupy the available space."""

        BOTH = tk.BOTH
        NONE = tk.NONE
        X = tk.X
        Y = tk.Y


class Grid(object, metaclass=ReadOnlyClass):
    """Grid geometry manager.

    https://www.pythontutorial.net/tkinter/tkinter-grid/
    """

    class Sticky(object, metaclass=ReadOnlyClass):
        """The sticky option specifies which edge of the cell the widget should stick to."""

        CENTER = tk.CENTER
        E = tk.E
        N = tk.N
        NE = tk.NE
        NW = tk.NW
        S = tk.S
        SE = tk.SE
        SW = tk.SW
        W = tk.W


class Place(object, metaclass=ReadOnlyClass):
    """Place geometry manager.

    https://www.pythontutorial.net/tkinter/tkinter-place/
    """

    class Anchor(object, metaclass=ReadOnlyClass):
        """The anchor parameter determines which part of the widget is positioned at the given coordinates."""

        CENTER = tk.CENTER
        E = tk.E
        N = tk.N
        NE = tk.NE
        NW = tk.NW
        S = tk.S
        SE = tk.SE
        SW = tk.SW
        W = tk.W


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
  Created: 29.10.2023

  Purpose: Class for creating and processes config files.
"""

from inspect import currentframe
from typing import List, Optional
from jsktoolbox.attribtool import NoDynamicAttributes, ReadOnlyClass
from jsktoolbox.raisetool import Raise
from jsktoolbox.libs.base_data import BData
from jsktoolbox.libs.system import PathChecker


class _Keys(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    FILE = "__file__"


class FileProcessor(BData, NoDynamicAttributes):
    """FileProcessor class."""

    def __init__(self) -> None:
        """Constructor."""

    @property
    def file(self) -> Optional[str]:
        """Return config file path."""
        if _Keys.FILE not in self._data:
            self._data[_Keys.FILE] = None
        if isinstance(self._data[_Keys.FILE], PathChecker):
            return self._data[_Keys.FILE].path
        return self._data[_Keys.FILE]

    @file.setter
    def file(self, path: str) -> None:
        """Set file name."""
        self._data[_Keys.FILE] = PathChecker(path)

    @property
    def file_exists(self) -> bool:
        """Check if the file exists and is a file."""
        obj: PathChecker = self._data[_Keys.FILE]
        return obj.exists and (obj.is_file or obj.is_symlink) and not obj.is_dir

    def file_create(self) -> bool:
        """Try to create file."""
        if self.file_exists:
            return True
        obj: PathChecker = self._data[_Keys.FILE]
        if obj.exists and obj.is_dir:
            raise Raise.error(
                f"Given path: {obj.path} exists and is a directory.",
                OSError,
                self._c_name,
                currentframe(),
            )
        return obj.create()

    def read(self) -> str:
        """Try to read config file."""
        out: str = ""
        if self.file_exists:
            filepath: Optional[str] = self.file
            if filepath is not None:
                with open(filepath, "r") as file:
                    out = file.read()
        return out

    def readlines(self) -> List[str]:
        """Try to read config file and create list of strings."""
        out: List[str] = []
        if self.file_exists:
            filepath: Optional[str] = self.file
            if filepath is not None:
                with open(filepath, "r") as file:
                    tmp = file.readlines()
                    for line in tmp:
                        if line.find("<end of section") > 0:
                            continue
                        out.append(line.strip())
        return out

    def write(self, data: str) -> None:
        """Try to write data to config file."""
        test: bool = self.file_exists
        if not test:
            test = self.file_create()
        if test:
            filepath: Optional[str] = self.file
            if filepath is not None:
                with open(filepath, "w") as file:
                    file.write(data)


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 16.10.2023

  Purpose: a simple class containing basic cryptographic procedures for strings.
"""

import string
from typing import Dict

from base64 import b64decode, b64encode
from codecs import getencoder
from inspect import currentframe
from random import randrange

from jsktoolbox.attribtool import NoDynamicAttributes
from jsktoolbox.raisetool import Raise

# https://www.tutorialspoint.com/cryptography_with_python/cryptography_with_python_xor_process.htm
# https://teachen.info/cspp/unit4/lab04-02.html


class SimpleCrypto(NoDynamicAttributes):
    """SimpleCrypto class.

    A class that allows performing simple cryptographic operations on strings of characters.
    """

    @staticmethod
    def chars_table_generator() -> str:
        """Return printable chars string.

        A static method that returns an extended string of printable characters
        that is used to generate a translation table in caesar methods.
        """
        return string.printable + "ĄĆĘŁŃÓŚŻŹąćęłńóśżź"

    @classmethod
    def salt_generator(cls, length: int = 8) -> int:
        """Method for generate random salt with specific length.

        length: int - number of digits in the generated salt

        Return: int - salt number
        """
        if length < 1:
            raise Raise.error(
                f"...{length}",
                ValueError,
                cls.__qualname__,
                currentframe(),
            )
        return randrange(int(10**length / 10), 10**length - 1)

    @classmethod
    def caesar_encrypt(cls, salt: int, message: str) -> str:
        """Caesar encoder method with chars translate table.

        salt: int    - a number used to calculate the offset in the translation table,
        message: str - string to encode,

        Return: str  - encoded string
        """
        if not isinstance(salt, int):
            raise Raise.error(
                "Expected 'salt' as integer.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        if not isinstance(message, str):
            raise Raise.error(
                "Expected 'message' as str type.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        chars: str = cls.chars_table_generator()
        chars_len: int = len(chars)
        shift: int = salt % chars_len
        transtable: Dict = str.maketrans(chars, chars[shift:] + chars[:shift])

        return message.translate(transtable)

    @classmethod
    def caesar_decrypt(cls, salt: int, message: str) -> str:
        """Caesar decoder method with chars translate table.

        salt: int    - a number used to calculate the offset in the translation table,
        message: str - encoded string,

        Return: str  - decoded string
        """
        if not isinstance(salt, int):
            raise Raise.error(
                "Expected 'salt' as integer.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        if not isinstance(message, str):
            raise Raise.error(
                "Expected 'message' as str type.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        chars: str = cls.chars_table_generator()
        chars_len: int = len(chars)
        shift: int = chars_len - (salt % chars_len)
        transtable: Dict = str.maketrans(chars, chars[shift:] + chars[:shift])

        return message.translate(transtable)

    @classmethod
    def rot13_codec(cls, message: str) -> str:
        """Rot13 encoder/decoder method.

        message: str - string for encode/decode

        Return: str - encoded/decoded string
        """
        if not isinstance(message, str):
            raise Raise.error(
                "Expected 'message' as str type.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        codec = lambda s: getencoder("rot13")(s)[0]
        return codec(message)  # type: ignore

    @classmethod
    def b64_encrypt(cls, message: str) -> str:
        """Base64 encoder method.

        message: str - string for encode,

        Return: str - base64 encoded string.
        """
        if not isinstance(message, str):
            raise Raise.error(
                "Expected 'message' as str type.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        return b64encode(message.encode("UTF-32")).decode()

    @classmethod
    def b64_decrypt(cls, message: str) -> str:
        """Base64 decoder method.

        message: str - base64 string for decode,

        Return: str - decoded string.
        """
        if not isinstance(message, str):
            raise Raise.error(
                "Expected 'message' as str type.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        return b64decode(message.encode("UTF-32")).decode("UTF-32")

    @classmethod
    def multiple_encrypt(cls, salt: int, message: str) -> str:
        """Multiple encoder method.

        salt: int    - a number used to calculate the offset in the translation table,
        message: str - string to encode,

        Return: str  - encoded string
        """
        if not isinstance(message, str):
            raise Raise.error(
                "Expected 'message' as str type.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        return cls.b64_encrypt(cls.caesar_encrypt(salt, cls.rot13_codec(message)))

    @classmethod
    def multiple_decrypt(cls, salt: int, message: str) -> str:
        """Multiple decoder method.

        salt: int    - a number used to calculate the offset in the translation table,
        message: str - encoded string,

        Return: str  - decoded string
        """
        if not isinstance(message, str):
            raise Raise.error(
                "Expected 'message' as str type.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        return cls.rot13_codec(cls.caesar_decrypt(salt, cls.b64_decrypt(message)))


# #[EOF]#######################################################################



# -*- coding: utf-8 -*-
"""
  base.py
  Author : Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
  Created: 15.01.2024, 13:41:50
  
  Purpose: Base classes for tkinter.
"""

from jsktoolbox.attribtool import NoDynamicAttributes, ReadOnlyClass


class TkBase(NoDynamicAttributes):
    """Base class for classess derived from Tk."""

    _name = None
    _tkloaded = None
    _w = None
    _windowingsystem_cached = None
    child = None
    children = None
    master = None
    tk = None
    widgetName = None


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
  Created: 29.10.2023

  Purpose: DataProcessor class for processing dataset operations.
"""

from inspect import currentframe
from typing import List, Tuple, Optional, Union, Any, TypeVar, Generic
from abc import ABC, abstractmethod
from copy import copy
from jsktoolbox.attribtool import NoDynamicAttributes, ReadOnlyClass
from jsktoolbox.raisetool import Raise
from jsktoolbox.libs.base_data import BData

TVariableModel = TypeVar("TVariableModel", bound="VariableModel")


class _Keys(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    DATA = "__data__"
    DESC = "__desc__"
    DESCRIPTION = "__description__"
    MAIN = "__main__"
    NAME = "__name__"
    VALUE = "__value__"
    VARIABLES = "__variables__"


class IModel(ABC):
    """Model class interface."""

    @property
    @abstractmethod
    def dump(self) -> Union[List[str], TVariableModel]:
        """Dump data."""

    @property
    @abstractmethod
    def name(self) -> Optional[str]:
        """Get name property."""

    @name.setter
    @abstractmethod
    def name(self, name: str) -> None:
        """Set name property."""

    @abstractmethod
    def parser(self, value: str) -> None:
        """Parser method."""

    @abstractmethod
    def search(self, name: str) -> bool:
        """Search method."""


class VariableModel(BData, IModel, NoDynamicAttributes):
    """VariableModel class."""

    def __init__(
        self,
        name: Optional[str] = None,
        value: Optional[Union[str, int, float, bool, List]] = None,
        desc: Optional[str] = None,
    ) -> None:
        """Constructor."""
        self._data[_Keys.NAME] = name
        self._data[_Keys.VALUE] = value
        self._data[_Keys.DESC] = desc

    def __repr__(self) -> str:
        """Return representation class string."""
        tmp: str = ""
        tmp += f"name='{self.name}', " if self.name is not None else ""
        if isinstance(self.value, (int, float, bool)):
            tmp += f"value={self.value}, " if self.value is not None else ""
        elif isinstance(self.value, List):
            tmp += f"value=[{self.value}], " if self.value is not None else ""
        else:
            tmp += f"value='{self.value}', " if self.value is not None else ""
        tmp += f"desc='{self.desc}'" if self.desc is not None else ""
        return f"{self._c_name}({tmp})"

    def __str__(self) -> str:
        """Return formated string."""
        tmp: str = ""
        tmp += f"{self.name} = " if self.name is not None else ""
        if isinstance(self.value, (int, float, bool)):
            tmp += f"{self.value}" if self.value is not None else ""
        elif isinstance(self.value, List):
            tmp += f"{self.value}" if self.value is not None else ""
        else:
            tmp += (
                '"{}"'.format(self.value.strip("\"'")) if self.value is not None else ""
            )
        if tmp:
            tmp += f" # {self.desc}" if self.desc is not None else ""
        else:
            tmp += f"# {self.desc}" if self.desc is not None else "#"
        return tmp

    @property
    def desc(self) -> Optional[str]:
        """Get descrption property."""
        return self._data[_Keys.DESC]

    @desc.setter
    def desc(self, desc: Optional[str]) -> None:
        """Set description property."""
        self._data[_Keys.DESC] = desc

    @property
    def dump(self) -> TVariableModel:
        """Dump data."""
        return self  # type: ignore

    @property
    def name(self) -> Optional[str]:
        """Get name property."""
        return self._data[_Keys.NAME]

    @name.setter
    def name(self, name: Optional[str]) -> None:
        """Set name property."""
        if name is None:
            return
        self._data[_Keys.NAME] = name.strip()

    def parser(self, value: str) -> None:
        """Parser method."""

    def search(self, name: str) -> bool:
        """Search method."""
        return self.name == name

    @property
    def value(self) -> Optional[Union[str, int, float, bool, List]]:
        """Get value property."""
        return self._data[_Keys.VALUE]

    @value.setter
    def value(self, value: Optional[Union[str, int, float, bool, List]]) -> None:
        """Set value property."""
        self._data[_Keys.VALUE] = value


class SectionModel(BData, IModel, NoDynamicAttributes):
    """SectionModel class."""

    def __init__(self, name: Optional[str] = None) -> None:
        """Constructor."""
        self._data[_Keys.NAME] = None
        self._data[_Keys.VARIABLES] = []
        self.parser(name)

    def __repr__(self) -> str:
        """Return representation class string."""
        return f"{self._c_name}(name='{self.name}')"

    def __str__(self) -> str:
        """Return formated string."""
        return f"[{self.name}]"

    @property
    def dump(self) -> List[Any]:
        """Dump data."""
        tmp: List = []
        tmp.append(self)
        for item in self._data[_Keys.VARIABLES]:
            tmp.append(item.dump())
        return copy(tmp)

    def parser(self, value: Optional[str]) -> None:
        """Parser method."""
        if value is None:
            return
        tmp: str = f"{value}".strip("[] \n")
        if tmp:
            self._data[_Keys.NAME] = tmp
        else:
            raise Raise.error(
                f"Expected String name, received: '{tmp}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def search(self, name: str) -> bool:
        """Search method."""
        return self.name == name

    @property
    def name(self) -> Optional[str]:
        """Get name property."""
        return self._data[_Keys.NAME]

    @name.setter
    def name(self, name: str) -> None:
        """Set name property."""
        self.parser(name)

    def get_variable(self, name: str) -> Optional[VariableModel]:
        """Search and return VariableModel if exists."""
        name = str(name)
        for item in self._data[_Keys.VARIABLES]:
            if item.name == name:
                return item
        return None

    def set_variable(
        self,
        name: Optional[str] = None,
        value: Optional[Any] = None,
        desc: Optional[str] = None,
    ) -> None:
        """Add or update VariableModel."""
        if name is not None:
            tmp: Optional[VariableModel] = self.get_variable(name)
            if tmp is not None:
                item: VariableModel = tmp
                if value is not None or (value is None and desc is None):
                    item.value = value
                if desc is not None or (desc is None and value is None):
                    item.desc = desc
                return
        # add new VariableModel
        if name is None and value is not None:
            return
        self._data[_Keys.VARIABLES].append(VariableModel(name, value, desc))

    @property
    def variables(self) -> List[VariableModel]:
        """Return list of VariableModel."""
        return self._data[_Keys.VARIABLES]


class DataProcessor(BData, NoDynamicAttributes):
    """DataProcessor class."""

    def __init__(self) -> None:
        """Constructor."""
        self._data[_Keys.DATA] = []

    @property
    def main_section(self) -> Optional[str]:
        """Return main section name."""
        if _Keys.MAIN not in self._data:
            self._data[_Keys.MAIN] = None
        return self._data[_Keys.MAIN]

    @main_section.setter
    def main_section(self, name: str) -> None:
        """Set main section name."""
        if not isinstance(name, str):
            name = str(name)
        self._data[_Keys.MAIN] = name
        self.add_section(name)

    @property
    def sections(self) -> Tuple:
        """Return sections keys tuple."""
        out = []
        for item in self._data[_Keys.DATA]:
            out.append(item.name)
        return tuple(sorted(out))

    def add_section(self, name: str) -> str:
        """Add section object to dataset.

        Return: extracted section name.
        """
        sm = SectionModel(str(name))
        if sm.name not in self.sections:
            self._data[_Keys.DATA].append(sm)
        return f"{sm.name}"

    def get_section(self, name: str) -> Optional[SectionModel]:
        """Get section object if exists."""
        sm = SectionModel(name)
        for item in self._data[_Keys.DATA]:
            if item.name == sm.name:
                return item
        return None

    def set(
        self,
        section: str,
        varname: Optional[str] = None,
        value: Optional[Any] = None,
        desc: Optional[str] = None,
    ) -> None:
        """Set data to [SectionModel]->[VariableModel]."""
        section_name: str = self.add_section(section)
        tmp: Optional[SectionModel] = self.get_section(section_name)
        if tmp is not None:
            found_section: SectionModel = tmp
            found_section.set_variable(varname, value, desc)

    def get(
        self, section: str, varname: Optional[str] = None, desc: bool = False
    ) -> Optional[Any]:
        """Return value."""
        sm = SectionModel(section)
        if sm.name in self.sections:
            tmp: Optional[SectionModel] = self.get_section(section)
            if tmp is not None:
                found_section: SectionModel = tmp
                if varname is not None:
                    found_var: Optional[VariableModel] = found_section.get_variable(
                        varname
                    )
                    if found_var is not None:
                        if desc:
                            # Return description for varname
                            return found_var.desc
                        else:
                            # Return value for varname
                            return found_var.value
                    else:
                        return None
                else:
                    # Return list od description for section
                    out: List[str] = []
                    for item in found_section.variables:
                        if item.name is None and item.desc is not None:
                            out.append(item.desc)
                    return out
        else:
            raise Raise.error(
                f"Given section name: '{section}' not found.",
                KeyError,
                self._c_name,
                currentframe(),
            )

    def __dump(self, section: str) -> str:
        """Return formatted configuration data for section name."""
        out: str = ""
        if section in self.sections:
            tmp: Optional[SectionModel] = self.get_section(section)
            if tmp is not None:
                found_section: SectionModel = tmp
                out += f"{found_section}\n"
                for item in found_section.variables:
                    out += f"{item}\n"
                out += f"# -----<end of section: '{found_section.name}'>-----\n"
        else:
            raise Raise.error(
                f"Section name: '{section}' not found.",
                KeyError,
                self._c_name,
                currentframe(),
            )
        return out

    @property
    def dump(self) -> str:
        """Return formated configuration data string."""
        out: str = ""

        # first section is a main section
        if self.main_section is None:
            raise Raise.error(
                "Main section is not set.",
                KeyError,
                self._c_name,
                currentframe(),
            )
        out = self.__dump(self.main_section)

        # other sections
        for section in sorted(tuple(set(self.sections) ^ set([self.main_section]))):
            out += self.__dump(section)

        return out


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
  Author:  Jacek Kotlarski --<szumak@virthost.pl>
  Created: 01.09.2023

  Purpose: BData container base class.
"""

from inspect import currentframe
from typing import Dict, Optional
from types import FrameType

from jsktoolbox.attribtool import NoDynamicAttributes
from jsktoolbox.raisetool import Raise


class BClasses(NoDynamicAttributes):
    """Base class for projects."""

    @property
    def _c_name(self) -> str:
        """Return class name."""
        return self.__class__.__name__

    @property
    def _f_name(self) -> str:
        """Return current method name."""
        tmp: Optional[FrameType] = currentframe()
        if tmp is not None:
            frame: Optional[FrameType] = tmp.f_back
            if frame is not None:
                method_name: str = frame.f_code.co_name
                return method_name
        return ""


class BData(BClasses):
    """BData container class."""

    __data: Optional[Dict] = None

    @property
    def _data(self) -> Dict:
        """Return data dict."""
        if self.__data is None:
            self.__data = {}
        return self.__data

    @_data.setter
    def _data(self, value: Optional[Dict]) -> None:
        """Set data dict."""
        if value is None:
            self.__data = {}
            return None
        if isinstance(value, Dict) and self.__data is not None:
            for key in value.keys():
                self.__data[key] = value[key]
        else:
            raise Raise.error(
                f"Expected Dict type, received: '{type(value)}'.",
                AttributeError,
                self._c_name,
                currentframe(),
            )


# #[EOF]#######################################################################
